"""module for fobj modifier."""
from __future__ import annotations
from typing import Any, Callable, cast, TYPE_CHECKING
from ..jobject import JObject
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class FObjModifier(Modifier):
    """Get object at field from a JSONClass object. If the object is not exist,
    include it and return it.
    """

    def __init__(self, field_name: str | Callable | Types) -> None:
        self.field_name = field_name

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        field_name = self.resolve_param(self.field_name, ctx)
        val = cast(JObject, ctx.val)
        if getattr(val, field_name) is None:
            val.include(field_name)
        return getattr(val, field_name)



"""module for tonextyear modifier."""
from __future__ import annotations
from datetime import datetime, date
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToNextYearModifier(Modifier):
    """The next first day of year after this date or datetime."""

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is datetime:
            return datetime(ctx.val.year + 1, 1, 1, 0, 0, 0, 0, tzinfo=ctx.val.tzinfo)
        if type(ctx.val) is date:
            return date(ctx.val.year + 1, 1, 1)
        else:
            return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleSecret:
    name: Optional[str] = types.str.trim
    message: Optional[str] = types.str.internal



"""module for setonsave modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class SetOnSaveModifier(Modifier):
    """Set on save modifier updates or sets value on save."""

    def __init__(self, setter: Callable | Types) -> None:
        from ..types import Types
        if not isinstance(setter, Types):
            if not callable(setter):
                raise ValueError('setonsave setter is not callable')
            params_len = len(signature(setter).parameters)
            if params_len > 1:
                raise ValueError('not a valid setonsave setter')
        self.setter = setter

    def serialize(self, ctx: Ctx) -> Any:
        if callable(self.setter):
            params_len = len(signature(self.setter).parameters)
            if params_len == 1:
                return self.setter(ctx.val)
            return self.setter()
        else:
            return self.setter.modifier.transform(ctx)



"""module for tostr modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ToStrModifier(Modifier):
    """ToStr Modifier transforms value into a str"""

    def transform(self, ctx: Ctx) -> Any:
        is_int_or_float_or_bool = type(ctx.val) is int or type(ctx.val) is float or type(ctx.val) is bool
        return str(ctx.val) if is_int_or_float_or_bool else ctx.val



"""module for listof modifier."""
from __future__ import annotations
from jsonclasses.vmsgcollector import VMsgCollector
from typing import (
    Any, Collection, Iterable, TypeVar, Union, TYPE_CHECKING
)
from ..fdef import FDef, Nullability
from ..jconf import JConf
from ..excs import ValidationException
from .type_modifier import TypeModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


T = TypeVar('T', bound=Collection)


class CollectionTypeModifier(TypeModifier):
    """Base class for collection type modifiers."""

    def __init__(self, raw_item_types: Any) -> None:
        super().__init__()
        self.raw_item_types = raw_item_types
        self.exact_type = False

    def define(self, fdef: FDef) -> None:
        super().define(fdef)
        fdef._raw_item_types = self.raw_item_types

    def enumerator(self, value: Collection) -> Iterable:
        raise NotImplementedError('please implement enumerator')

    def empty_collection(self) -> Collection:
        raise NotImplementedError('please override empty_collection')

    def append_value(self, i: Union[str, int], v: Any, col: Collection):
        raise NotImplementedError('please implement append_value')

    def to_object_key(self, key: T, conf: JConf) -> T:
        return key

    def to_json_key(self, key: T, conf: JConf) -> T:
        return key

    def should_special_handle(self, key: Any, v: Any, ctx: Ctx) -> bool:
        return False

    def special_handle(self, key: Any, v: Any, ctx: Ctx) -> None:
        pass

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        super().validate(ctx)
        itypes = ctx.fdef.item_types
        all_fields = next(b for b in [
            ctx.ctxcfg.all_fields,
            ctx.cdefowner.jconf.validate_all_fields] if b is not None)
        vmsgctor = VMsgCollector()
        for i, v in self.enumerator(ctx.val):
            try:
                ictx = ctx.colval(v, i, itypes.fdef, ctx.val)
                itypes.modifier.validate(ictx)
            except ValidationException as exception:
                if all_fields:
                    vmsgctor.receive(exception.keypath_messages)
                else:
                    raise exception
        if vmsgctor.has_msgs:
            ctx.raise_mvexc(vmsgctor.messages)

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            if ctx.fdef.collection_nullability == Nullability.NONNULL:
                return self.empty_collection()
            else:
                return None
        if not isinstance(ctx.val, self.cls):
            return ctx.val
        itypes = ctx.fdef.item_types
        retval = self.empty_collection()
        for i, v in self.enumerator(ctx.val):
            if self.should_special_handle(i, v, ctx):
                self.special_handle(i, v, ctx)
            else:
                ictx = ctx.colval(v, i, itypes.fdef, ctx.val)
                tsfmd = itypes.modifier.transform(ictx)
                self.append_value(
                    self.to_object_key(i, ctx.owner.cdef.jconf), tsfmd, retval)
        return retval

    def tojson(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if not isinstance(ctx.val, self.cls):
            return ctx.val
        itypes = ctx.fdef.item_types
        retval = self.empty_collection()
        for i, v in self.enumerator(ctx.val):
            ictx = ctx.colval(v, i, itypes.fdef, ctx.val)
            tsfmd = itypes.modifier.tojson(ictx)
            self.append_value(
                    self.to_json_key(i, ctx.owner.cdef.jconf), tsfmd, retval)
        return retval

    def serialize(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if not isinstance(ctx.val, self.cls):
            return ctx.val
        itypes = ctx.fdef.item_types
        retval = self.empty_collection()
        for i, v in self.enumerator(ctx.val):
            ictx = ctx.colval(v, i, itypes.fdef, ctx.val)
            tsfmd = itypes.modifier.serialize(ictx)
            self.append_value(
                    self.to_json_key(i, ctx.cdefowner.jconf), tsfmd, retval)
        return retval



"""module for isobjof modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..jobject import JObject

class IsObjOfModifier(Modifier):
    """Value if value is object of class."""

    def __init__(self, cls: type[JObject] | str) -> None:
        self.cls = cls

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            ctx.raise_vexc('none is not object of a class')
        if type(self.cls) is str:
            self.cls = ctx.owner.__class__.cdef.jconf.cgraph.fetch(self.cls).cls
        if not isinstance(ctx.val, self.cls):
            ctx.raise_vexc(f'value is not instance of {self.cls.__name__}')



"""module for checkpw modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class CheckpwModifier(Modifier):
    """Checkpw modifier checks string value with bcrypt's checkpw function."""

    def __init__(self, against: Types) -> None:
        self.against = against
        self.check_packages()

    def packages(self) -> dict[str, (str, str)] | None:
        return {'bcrypt': ('bcrypt', '>=3.2.0,<4.0.0')}

    def validate(self, ctx: Ctx) -> None:
        from bcrypt import checkpw
        against_val = self.against.modifier.transform(ctx)
        if type(against_val) is not str:
            ctx.raise_vexc('value is incorrect')
        if type(ctx.val) is not str:
            ctx.raise_vexc('value is incorrect')
        if not checkpw(against_val.encode(), ctx.val.encode()):
            ctx.raise_vexc('value is incorrect')



from __future__ import annotations
from tests.classes.super_str import SuperStr
from unittest import TestCase



class TestPadEnd(TestCase):

    def test_padend_adds_str_to_the_end_of_str(self):
        ps = SuperStr(pade="aaaaa")
        self.assertEqual(ps.pade, "aaaaaeeeee")

    def test_padend_keeps_value_when_type_of_value_is_not_str(self):
        ps = SuperStr(padie=80)
        self.assertEqual(ps.padie, 80)

    def test_padend_keeps_value_when_target_length_less_value_length(self):
        ps = SuperStr(pade="aaaaaaaaaaa")
        self.assertEqual(ps.pade, "aaaaaaaaaaa")

    def test_padend_adds_callable_str_to_the_end_of_str(self):
        ps = SuperStr(padce="aaaaa")
        self.assertEqual(ps.padce, "aaaaaeeeee")

    def test_padend_adds_types_str_to_the_end_of_str(self):
        ps = SuperStr(padte="aaaaa")
        self.assertEqual(ps.padte, "aaaaaeeeee")



from __future__ import annotations
from tests.classes.simple_math import SimpleMath
from jsonclasses.excs import ValidationException
from unittest import TestCase


class TestSqrt(TestCase):

    def test_sqrt_with_get_the_square_root_of_int_value(self):
        s = SimpleMath(i_sqrt=4)
        self.assertEqual(s.i_sqrt, 2)

    def test_sqrt_with_get_the_square_root_of_float_value(self):
        s = SimpleMath(f_sqrt=6.25)
        self.assertEqual(s.f_sqrt, 2.5)

    def test_sqrt_raises_if_value_less_or_equal_to_zero(self):
        with self.assertRaises(ValidationException) as context:
            s = SimpleMath(f_sqrt=-6.25)
        self.assertEqual(context.exception.keypath_messages['fSqrt'],
                         "value is less than 0 thus cannot be sqrted")



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleCalculation:

    i_add: Optional[int] = types.int.add(5)
    f_add: Optional[float] = types.float.add(2.5)
    c_add: Optional[float] = types.float.add(lambda: 2.5)
    t_add: Optional[float] = types.float.add(types.default(2.5))

    i_sub: Optional[int] = types.int.sub(5)
    f_sub: Optional[float] = types.float.sub(2.5)
    c_sub: Optional[float] = types.float.sub(lambda: 2.5)
    t_sub: Optional[float] = types.float.sub(types.default(2.5))

    i_mul: Optional[int] = types.int.mul(5)
    f_mul: Optional[float] = types.float.mul(2.5)
    c_mul: Optional[float] = types.float.mul(lambda: 2.5)
    t_mul: Optional[float] = types.float.mul(types.default(2.5))

    i_div: Optional[int] = types.int.div(5)
    f_div: Optional[float] = types.float.div(2.5)
    c_div: Optional[float] = types.float.div(lambda: 4.2)
    t_div: Optional[float] = types.float.div(types.default(4.2))

    i_mod: Optional[int] = types.int.mod(5)
    f_mod: Optional[float] = types.float.mod(2.5)
    c_mod: Optional[float] = types.float.mod(lambda: 3.5)
    t_mod: Optional[float] = types.float.mod(types.default(3.5))



from __future__ import annotations
from typing import TYPE_CHECKING
from jsonclasses import jsonclass, types
if TYPE_CHECKING:
    from tests.classes.author import Author


@jsonclass(strict_input=False, validate_all_fields=True)
class Article:
    title: str
    content: str
    author: Author = types.objof('Author').linkto.required



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_int import SuperInt


class TestEven(TestCase):

    def test_even_doesnt_raise_if_int_value_is_even(self):
        i = SuperInt(i_e=6)
        self.assertEqual(i.i_e, 6)

    def test_even_raises_if_int_value_is_not_even(self):
        i = SuperInt(i_e=3)
        with self.assertRaises(ValidationException) as context:
            i.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['iE'],
                         "value is not even")



"""This module defines JSON Class context objects."""
from __future__ import annotations
from typing import Any, NamedTuple, Optional, cast, TYPE_CHECKING
from .jconf import JConf
from .types import types
from .mgraph import MGraph
from .excs import ValidationException
if TYPE_CHECKING:
    from .cdef import CDef
    from .fdef import FDef
    from .jobject import JObject


class CtxCfg(NamedTuple):

    all_fields: Optional[bool] = None
    """On validating, whether validate all fields.
    """

    ignore_writeonly: Optional[bool] = None
    """On tojson, whether ignore writeonly.
    """

    reverse_relationship: Optional[bool] = None
    """On tojson, whether output reverse relationship.
    """

    fill_dest_blanks: Optional[bool] = None
    """On set and init, whether fill default fields with None value. This is an
    internal behavior. Do not set this.
    """

    output_null: Optional[bool] = None
    """On tojson, whether output null value instead of unexisting field.
    """

class Ctx(NamedTuple):
    root: JObject
    owner: JObject
    parent: list | dict | JObject
    holder: Optional[JObject]
    val: Any
    original: Any
    ctxcfg: CtxCfg
    keypathr: list[str | int]
    fkeypathr: list[str | int]
    keypatho: list[str | int]
    fkeypatho: list[str | int]
    keypathp: list[str | int]
    fkeypathp: list[str | int]
    keypathh: list[str | int]
    fkeypathh: list[str | int]
    fdef: FDef
    operator: Any
    mgraph: MGraph = MGraph()
    idchain: list[str] = []
    passin: Optional[Any] = None

    @property
    def cdefroot(self: Ctx) -> CDef:
        return self.root.__class__.cdef

    @property
    def cdefowner(self: Ctx) -> CDef:
        return self.owner.__class__.cdef

    @property
    def cdefholder(self: Ctx) -> CDef:
        return cast(JObject, self.holder).__class__.cdef

    @property
    def jconfroot(self: Ctx) -> JConf:
        return self.cdefroot.jconf

    @property
    def jconfowner(self: Ctx) -> JConf:
        return self.cdefowner.jconf

    @property
    def jconfholder(self: Ctx) -> JConf:
        return self.cdefholder.jconf

    @property
    def iscreate(self: Ctx) -> bool:
        return self.original is None

    @property
    def skeypathr(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.keypathr])

    @property
    def skeypatho(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.keypatho])

    @property
    def skeypathp(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.keypathp])

    @property
    def skeypathh(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.keypathh])

    @property
    def sfkeypathr(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.fkeypathr])

    @property
    def sfkeypatho(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.fkeypatho])

    @property
    def sfkeypathp(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.fkeypathp])

    @property
    def sfkeypathh(self: Ctx) -> str:
        return '.'.join([str(k) for k in self.fkeypathh])

    @classmethod
    def rootctx(cls: type[Ctx], root: JObject, ctxcfg: CtxCfg,
                value: Any = None) -> Ctx:
        fdef = types.objof(root.__class__).fdef
        fdef._cdef = root.__class__.cdef
        return Ctx(root=root, owner=root, parent=root, holder=None,
                   val=value if value is not None else root,
                   original=root, ctxcfg=ctxcfg, keypatho=[], fkeypatho=[],
                   keypathr=[], fkeypathr=[], keypathp=[], fkeypathp=[],
                   keypathh=[], fkeypathh=[], fdef=fdef,
                   operator=root._operator, mgraph=MGraph(), idchain=[])

    @classmethod
    def rootctxp(cls: type[Ctx], root: JObject, key: str, val: Any, passin: Any) -> Ctx:
        fdef = types.objof(root.__class__).fdef
        fdef._cdef = root.__class__.cdef
        ekey = root.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=root, owner=root, parent=root, holder=None, val=val,
                   original=root, ctxcfg=CtxCfg(),
                   keypatho=[key], fkeypatho=[ekey],
                   keypathr=[key], fkeypathr=[ekey],
                   keypathp=[key], fkeypathp=[ekey],
                   keypathh=[key], fkeypathh=[ekey],
                   fdef=fdef,
                   operator=root._operator, mgraph=MGraph(), idchain=[],
                   passin=passin)

    def alterfdef(self: Ctx, fdef: FDef) -> Ctx:
        return Ctx(root=self.root, owner=self.owner, parent=self.parent, holder=self.holder,
                   val=self.val,
                   original=self.original, ctxcfg=self.ctxcfg, keypatho=self.keypatho, fkeypatho=self.fkeypatho,
                   keypathr=self.keypathr, fkeypathr=self.fkeypathr, keypathp=self.keypathp, fkeypathp=self.fkeypathp,
                   keypathh=self.keypathh, fkeypathh=self.fkeypathh, fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph, idchain=self.idchain)

    def nval(self: Ctx, newval: Any) -> Ctx:
        return Ctx(root=self.root, owner=self.owner, parent=self.parent,
                   holder=self.holder, val=newval, original=self.original,
                   ctxcfg=self.ctxcfg,
                   keypatho=self.keypatho, fkeypatho=self.fkeypatho,
                   keypathr=self.keypathr, fkeypathr=self.fkeypathr,
                   keypathp=self.keypathp, fkeypathp=self.fkeypathp,
                   keypathh=self.keypathh, fkeypathh=self.fkeypathh,
                   fdef=self.fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=self.idchain, passin=self.passin)

    def nextv(self: Ctx, val: Any, key: str | int, fdef: FDef) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=self.owner, parent=self.parent,
                   holder=self.holder, val=val, original=None,
                   ctxcfg=self.ctxcfg,
                   keypatho=[*self.keypatho, key],
                   fkeypatho=[*self.fkeypatho, ekey],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[*self.keypathp, key],
                   fkeypathp=[*self.fkeypathp, ekey],
                   keypathh=[*self.keypathh, key],
                   fkeypathh=[*self.fkeypathh, ekey],
                   fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=self.idchain, passin=self.passin)

    def nexto(self: Ctx, val: Any, key: str | int, fdef: FDef) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=val, parent=val, holder=self.owner,
                   val=val, original=None, ctxcfg=self.ctxcfg, keypatho=[],
                   fkeypatho=[],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[], fkeypathp=[],
                   keypathh=[key], fkeypathh=[ekey], fdef=fdef,
                   operator=self.operator,
                   mgraph=self.mgraph, idchain=self.idchain,
                   passin=self.passin)

    def nextvc(self: Ctx, val: Any, key: str | int, fdef: FDef, c: str) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=self.owner, parent=self.parent,
                   holder=self.holder, val=val, original=None,
                   ctxcfg=self.ctxcfg,
                   keypatho=[*self.keypatho, key],
                   fkeypatho=[*self.fkeypatho, ekey],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[*self.keypathp, key],
                   fkeypathp=[*self.fkeypathp, ekey],
                   keypathh=[*self.keypathh, key],
                   fkeypathh=[*self.fkeypathh, ekey],
                   fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=[*self.idchain, c], passin=self.passin)

    def nextoc(self: Ctx, val: Any, key: str | int, fdef: FDef, c: str) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=val, parent=val, holder=self.owner,
                   val=val, original=None, ctxcfg=self.ctxcfg,
                   keypatho=[],
                   fkeypatho=[],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[],
                   fkeypathp=[],
                   keypathh=[key],
                   fkeypathh=[ekey],
                   fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=[*self.idchain, c], passin=self.passin)

    def nextvo(self: Ctx, val: Any, key: str | int, fdef: FDef, o: JObject) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=o, parent=self.parent,
                   holder=self.holder, val=val, original=None,
                   ctxcfg=self.ctxcfg,
                   keypatho=[*self.keypatho, key],
                   fkeypatho=[*self.fkeypatho, ekey],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[*self.keypathp, key],
                   fkeypathp=[*self.fkeypathp, ekey],
                   keypathh=[*self.keypathh, key],
                   fkeypathh=[*self.fkeypathh, ekey],
                   fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=self.idchain, passin=self.passin)

    def colval(self: Ctx, val: Any, key: str | int, fdef: FDef, p: Any) -> Ctx:
        return Ctx(root=self.root, owner=self.owner, parent=p,
                   holder=self.holder,
                   val=val, original=None, ctxcfg=self.ctxcfg,
                   keypatho=[*self.keypatho, key],
                   fkeypatho=[*self.fkeypatho, key],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, key],
                   keypathp=[key],
                   fkeypathp=[*self.fkeypathp, key],
                   keypathh=[*self.keypathh, key],
                   fkeypathh=[*self.fkeypathh, key],
                   fdef=fdef,
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=self.idchain, passin=self.passin)

    def default(self: Ctx, owner: JObject, key: str | int, fdef: FDef) -> Ctx:
        ekey = self.owner.__class__.cdef.jconf.output_key_strategy(key)
        return Ctx(root=self.root, owner=owner, parent=owner,
                   holder=self.holder, val=None,
                   original=None, ctxcfg=self.ctxcfg,
                   keypatho=[*self.keypatho, key],
                   fkeypatho=[*self.fkeypatho, ekey],
                   keypathr=[*self.keypathr, key],
                   fkeypathr=[*self.fkeypathr, ekey],
                   keypathp=[*self.keypathp, key],
                   fkeypathp=[*self.fkeypathp, ekey],
                   keypathh=[*self.keypathh, key], fdef=fdef,
                   fkeypathh=[*self.fkeypathh, ekey],
                   operator=self.operator, mgraph=self.mgraph,
                   idchain=self.idchain, passin=self.passin)

    def raise_vexc(self: Ctx, msg: str) -> None:
        """Raise validation error with message.
        """
        raise ValidationException({self.sfkeypathr: msg}, self.root)

    def raise_mvexc(self: Ctx, msgs: dict[str, str]) -> None:
        raise ValidationException(msgs, self.root)



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_str import SuperStr


class TestSecurepw(TestCase):

    def test_securepw_does_not_raise_if_value_is_scure_password(self):
        pw = SuperStr(securepw='w4Hsa/')
        pw.validate()

    def test_securepw_raises_if_value_is_not_scure_password(self):
        pw = SuperStr(securepw='w4Hsa')
        with self.assertRaises(ValidationException) as context:
             pw.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['securepw'],
                         "value is not secure password")



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.zero_number import ZeroNumber


class TestNegative(TestCase):

    def test_negative_doesnt_raise_if_float_value_is_less_than_zero(self):
        n = ZeroNumber(fnegative=-5.5)
        n.validate()

    def test_negative_raises_if_float_value_is_equal_to_zero(self):
        n = ZeroNumber(fnegative=0)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['fnegative'],
                         "value is not negative")

    def test_negative_raises_if_float_value_is_greater_than_zero(self):
        n = ZeroNumber(fnegative=5.5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['fnegative'],
                         "value is not negative")

    def test_negative_doesnt_raise_if_int_value_is_less_than_zero(self):
        n = ZeroNumber(inegative=-5)
        n.validate()

    def test_negative_raises_if_int_value_is_equal_to_zero(self):
        n = ZeroNumber(inegative=0)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['inegative'],
                         "value is not negative")

    def test_negative_raises_if_int_value_is_greater_than_zero(self):
        n = ZeroNumber(inegative=5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['inegative'],
                         "value is not negative")



"""module for abs modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class AbsModifier(Modifier):
    """Abs modifier transforms number value to its absolute value."""

    def transform(self, ctx: Ctx) -> Any:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        return abs(ctx.val) if is_number else ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_permission import (
    SuperPermissionArticle, SuperPermissionUser
)


class TestCanCUR(TestCase):

    def test_canc_wont_allow_if_operator_is_not_present(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=504, user=user, title='1', content='2')
        with self.assertRaises(ValidationException) as context:
            article.save()
        self.assertEqual(context.exception.keypath_messages, {
            'title': 'operator not present'
        })

    def test_canc_wont_allow_if_doesnt_pass_canc_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=504, user=user, title='1', content='2')
        article.opby(user)
        with self.assertRaises(ValidationException) as context:
            article.save()
        self.assertEqual(context.exception.keypath_messages, {
            'title': 'operation is not permitted'
        })

    def test_canc_allows_if_passes_canc_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=505, user=user, title='1', content='2')
        article.opby(user)
        article.save()

    def test_canu_wont_allow_if_operator_is_not_present(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=504, user=user, title='1', content='2', content2='3')
        setattr(article, "_is_new", False)
        article.content2 = '4'
        with self.assertRaises(ValidationException) as context:
            article.save(validate_all_fields=True)
        self.assertEqual(context.exception.keypath_messages, {
            'content2': 'operator not present'
        })

    def test_canu_wont_allow_if_doesnt_pass_canu_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        user2 = SuperPermissionUser(id=2, code=500)
        article = SuperPermissionArticle(id=2, code=505, user=user2, title='1', content='2', content2='3')
        setattr(article, "_is_new", False)
        article.content2 = '4'
        article.opby(user)
        with self.assertRaises(ValidationException) as context:
            article.save()
        self.assertEqual(context.exception.keypath_messages, {
            'content2': 'operation is not permitted'
        })

    def test_canu_allows_if_passes_canu_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=505, user=user, title='1', content='2', content2='3')
        setattr(article, "_is_new", False)
        article.content2 = '4'
        article.opby(user)
        article.save()

    def test_canr_remove_field_if_operator_is_not_present(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=504, user=user, title='1', content='2')
        json = article.tojson()
        self.assertNotIn('content', json)

    def test_canr_remove_field_if_doesnt_pass_canr_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=504, user=user, title='1', content='2')
        operator = SuperPermissionUser(id=1, code=505)
        article.opby(operator)
        json = article.tojson()
        self.assertNotIn('content', json)

    def test_canu_outputs_value_if_passes_canu_checker(self):
        user = SuperPermissionUser(id=1, code=505)
        article = SuperPermissionArticle(id=2, code=505, user=user, title='1', content='2')
        article.opby(user)
        json = article.tojson()
        self.assertEqual(json['content'], '2')



"""module for readonly modifier."""
from ..fdef import FDef
from .modifier import Modifier


class PrimaryModifier(Modifier):
    """Primary modifier marks a field as the primary key."""

    def define(self, fdef: FDef) -> None:
        fdef._primary = True



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.super_number import SuperNumber

class TestMax(TestCase):

    def test_max_doesnt_raise_if_float_value_is_less_than_arg(self):
        n = SuperNumber(ff_max=5.4)
        n.validate()

    def test_max_doesnt_raise_if_float_value_is_equal_to_arg(self):
        n = SuperNumber(ff_max=5.5)
        n.validate()

    def test_max_raises_if_float_value_is_greater_than_arg(self):
        n = SuperNumber(ff_max=5.6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffMax'],
                         "value is not less than or equal 5.5")

    def test_max_doesnt_raise_if_int_value_is_less_than_arg(self):
        n = SuperNumber(if_max=4)
        n.validate()

    def test_max_doesnt_raise_if_int_value_is_equal_to_arg(self):
        n = SuperNumber(if_max=5)
        n.validate()

    def test_max_raises_if_int_value_is_greater_than_arg(self):
        n = SuperNumber(if_max=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifMax'],
                         "value is not less than or equal 5")

    def test_max_doesnt_raise_if_callable_value_is_less_than_arg(self):
        n = SuperNumber(c_max=4)
        n.validate()

    def test_max_doesnt_raise_if_callable_value_is_equal_to_arg(self):
        n = SuperNumber(c_max=5)
        n.validate()

    def test_max_raises_if_callable_value_is_greater_than_arg(self):
        n = SuperNumber(c_max=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['cMax'],
                         "value is not less than or equal 5")

    def test_max_doesnt_raise_if_types_value_is_less_than_arg(self):
        n = SuperNumber(t_max=4)
        n.validate()

    def test_max_doesnt_raise_if_types_value_is_equal_to_arg(self):
        n = SuperNumber(t_max=5)
        n.validate()

    def test_max_raises_if_types_value_is_greater_than_arg(self):
        n = SuperNumber(t_max=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['tMax'],
                         "value is not less than or equal 5")



"""module for referee modifier."""
from ..fdef import FDef
from .modifier import Modifier


class RefereeModifier(Modifier):
    """Readwrite modifier marks a reference field with a referee name."""

    def __init__(self, referee_key: str) -> None:
        self.referee_key = referee_key

    def define(self, fdef: FDef) -> None:
        fdef._join_table_referee_key = self.referee_key



"""module for salt modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class SaltModifier(Modifier):
    """Salt modifier add salt to a string. This is mostly used for password."""

    def __init__(self) -> None:
        self.check_packages()

    def packages(self) -> dict[str, (str, str)] | None:
        return {'bcrypt': ('bcrypt', '>=3.2.0,<4.0.0')}

    def transform(self, ctx: Ctx) -> Any:
        from bcrypt import hashpw, gensalt
        if type(ctx.val) is str:
            return hashpw(ctx.val.encode(), gensalt()).decode('utf-8')
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_suffix import SuperSuffix


class TestHasSuffix(TestCase):

    def test_hassuffix_validates_str_param_is_suffix_of_str_val(self):
        s_hs = SuperSuffix(s_hs='this is python')
        s_hs.validate()

    def test_hassuffix_validates_list_param_is_suffix_of_list_val(self):
        l_hs = SuperSuffix(l_hs=['qq', 'dd', 'ee', 'ff'])
        l_hs.validate()

    def test_hassuffix_raises_if_str_param_is_not_suffix_of_str_val(self):
        s_hs = SuperSuffix(s_hs='This ')
        with self.assertRaises(ValidationException) as context:
            s_hs.validate()
        self.assertEqual(context.exception.keypath_messages['sHs'],
                         "suffix is not found")

    def test_hassuffix_raises_if_list_param_is_not_suffix_of_list_val(self):
        l_hs = SuperSuffix(l_hs=['qq'])
        with self.assertRaises(ValidationException) as context:
            l_hs.validate()
        self.assertEqual(context.exception.keypath_messages['lHs'],
                         "suffix is not found")

    def test_hassuffix_validates_callable_param_is_suffix_of_str_val(self):
        c_hs = SuperSuffix(c_hs='this is python')
        c_hs.validate()

    def test_hassuffix_validates_callable_param_is_suffix_of_list_val(self):
        c_l_hs = SuperSuffix(c_l_hs=['qq', 'dd', 'ee', 'ff'])
        c_l_hs.validate()

    def test_hassuffix_raises_if_callable_param_is_not_suffix_of_str_val(self):
        c_hs = SuperSuffix(c_hs='This ')
        with self.assertRaises(ValidationException) as context:
            c_hs.validate()
        self.assertEqual(context.exception.keypath_messages['cHs'],
                         "suffix is not found")

    def test_hassuffix_raises_if_callable_param_is_not_suffix_of_list_val(self):
        c_l_hs = SuperSuffix(c_l_hs=['qq'])
        with self.assertRaises(ValidationException) as context:
            c_l_hs.validate()
        self.assertEqual(context.exception.keypath_messages['cLHs'],
                         "suffix is not found")

    def test_hassuffix_validates_types_param_is_suffix_of_str_val(self):
        t_hs = SuperSuffix(t_hs='this is python')
        t_hs.validate()

    def test_hassuffix_validates_types_param_is_suffix_of_list_val(self):
        t_l_hs = SuperSuffix(t_l_hs=['qq', 'dd', 'ee', 'ff'])
        t_l_hs.validate()

    def test_hassuffix_raises_if_types_param_is_not_suffix_of_str_val(self):
        t_hs = SuperSuffix(t_hs='This ')
        with self.assertRaises(ValidationException) as context:
            t_hs.validate()
        self.assertEqual(context.exception.keypath_messages['tHs'],
                         "suffix is not found")

    def test_hassuffix_raises_if_types_param_is_not_suffix_of_list_val(self):
        t_l_hs = SuperSuffix(t_l_hs=['qq'])
        with self.assertRaises(ValidationException) as context:
            t_l_hs.validate()
        self.assertEqual(context.exception.keypath_messages['tLHs'],
                         "suffix is not found")



"""
This module contains `jsonenum`, the decorator for enums used with jsonclasses.
"""
from typing import Optional, Union, Callable, TypeVar, overload
from enum import Enum
from .cgraph import CGraph


T = TypeVar('T', bound=type[Enum])

@overload
def jsonenum(cls: T) -> T: ...


@overload
def jsonenum(
    cls: None,
    class_graph: Optional[str] = 'default'
) -> Callable[[T], T]: ...


@overload
def jsonenum(
    cls: type,
    class_graph: Optional[str] = 'default'
) -> T: ...


def jsonenum(
    cls: Optional[type] = None,
    class_graph: Optional[str] = 'default',
) -> Union[Callable[[T], T], T]:
    """The jsonclass enum decorator. To declare a jsonclass enum, use this
    syntax:

      @jsonenum
      class MyColor(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3
    """
    if cls is not None:
        if not isinstance(cls, type):
            raise ValueError('@jsonenum should be used to decorate a class.')
        cgraph = CGraph(class_graph)
        cgraph.put_enum(cls)
        return cls
    else:
        def parametered_jsonenum(cls):
            return jsonenum(
                cls,
                class_graph=class_graph)
        return parametered_jsonenum



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_map import SuperMap


class TestMap(TestCase):

    def test_map_maps_value_by_callable(self):
        item = SuperMap(list1=[0, 1, 2, 3, 4])
        self.assertEqual(item.list1, [1, 2, 3, 4, 5])

    def test_map_maps_value_by_types(self):
        item = SuperMap(list2=[0, 1, 2, 3, 4])
        self.assertEqual(item.list2, [1, 2, 3, 4, 5])



"""
JSON Classes
~~~~~~~~~~~~

JSON Classes is the Modern Declarative Data Flow and Data Graph Framework for
the AI Empowered Generation.

:copyright: (c) 2020 by Fillmula Inc., Victor Teo

:license: MIT, see LICENSE for more details.
"""
# flake8: noqa: F401
from .jsonclass import jsonclass
from .jsonenum import jsonenum
from .types import types
from .typing import linkto, linkedby, linkedthru
from .encoder import JSONEncoder
from .isjsonclass import isjsonclass, isjsonobject
from .uconf import uconf



from __future__ import annotations
from tests.classes.super_iterable import SuperIterable
from unittest import TestCase



class TestReplace(TestCase):

    def test_replace_keeps_str_value_if_no_replacement_is_found(self):
        s = SuperIterable(itsrp="bbbccc")
        self.assertEqual(s.itsrp, "bbbccc")

    def test_replace_replaces_first_appearance(self):
        s = SuperIterable(itsrp="dd qq abc rr")
        self.assertEqual(s.itsrp, "dd qq ABC rr")

    def test_replace_replaces_multiple_appearances(self):
        s = SuperIterable(itsrp="gg ss abc dd abc ee abc")
        self.assertEqual(s.itsrp, "gg ss ABC dd ABC ee ABC")

    def test_replace_replaces_first_appearance_with_callable(self):
        s = SuperIterable(itsrpc="dd qq abc rr")
        self.assertEqual(s.itsrpc, "dd qq ABC rr")



from __future__ import annotations
from unittest import TestCase
from tests.classes.linked_profile import LinkedProfile
from tests.classes.linked_user import LinkedUser


class TestSerialize(TestCase):

    def test_serialize_linked_objects_wont_go_into_infinite_loop(self):
        profile = LinkedProfile(name='Ua Bê Tshiu Pên')
        user = LinkedUser(name='Tsuan Sê Kai')
        user.profile = profile
        user._set_on_save()



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass


@jsonclass
class NestedList:
    id: Optional[str]
    value: dict[str, list[str]]


@jsonclass
class NestedDict:
    id: Optional[str]
    value: list[dict[str, str]]



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.product_with_numeric import NumericProductId


class TestDigit(TestCase):

    def test_numeric_doesnt_raise_if_value_is_numeric(self):
        analysis = NumericProductId(product_name='water', product_id='12345')
        analysis.validate()

    def test_numeric_doesnt_raise_if_value_is_numeric_in_unicode(self):
        analysis = NumericProductId(product_name='water', product_id='\u00BC')
        analysis.validate()

    def test_numeric_raises_if_value_is_float(self):
        analysis = NumericProductId(product_name='water', product_id='12.1')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productId'],
                         "value is not numeric string")

    def test_numeric_raises_if_value_contains_alphabet(self):
        analysis = NumericProductId(product_name='water', product_id='12a')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productId'],
                         "value is not numeric string")

    def test_numeric_raises_if_value_contains_special_characters(self):
        analysis = NumericProductId(product_name='water', product_id='12!')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productId'],
                         "value is not numeric string")



from __future__ import annotations
from datetime import datetime, timezone
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_mixed import SimpleMixed, SimpleMixedT, SimpleMixedU


class TestUnion(TestCase):

    def test_union_raises_if_none_matches(self):
        mixed = SimpleMixed(name='Mixed', mixed=5.0)
        with self.assertRaises(ValidationException):
            mixed.validate()

    def test_union_does_not_raise_if_any_matches(self):
        mixed = SimpleMixed(name='Mixed', mixed=5)
        mixed.validate()
        mixed1 = SimpleMixed(name='Mixed', mixed='5')
        mixed1.validate()

    def test_union_transform_use_matched_transform(self):
        one = SimpleMixedT(value=50)
        self.assertEqual(one.value, 173)
        two = SimpleMixedT(value='50')
        self.assertEqual(two.value, '50123')

    def test_union_tojson_use_matched_displayer(self):
        one = SimpleMixedU(value='2021-10-16T13:43:28.099Z')
        self.assertEqual(one.value, datetime(2021, 10, 16, 13, 43, 28, 99000, tzinfo=timezone.utc))
        two = SimpleMixedU(value='123')
        self.assertEqual(two.value, '123')
        self.assertEqual(one.tojson()['value'], '2021-10-16T13:43:28.099Z')
        self.assertEqual(two.tojson()['value'], '123')



"""module for eager modifier."""
from .modifier import Modifier
from ..fdef import FDef


class EagerModifier(Modifier):
    """An EagerModifier marks fields for initialization and set stage validation.
    This is used usually before heavy transforming modifiers.
    """

    def define(self, fdef: FDef) -> None:
        fdef._has_eager_modifier = True



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_type import SuperType


class TestToStr(TestCase):

    def test_tostr_transforms_int_into_str(self):
        s = SuperType(ts=-122)
        self.assertEqual(s.ts, "-122")

    def test_tostr_transforms_float_into_str(self):
        s = SuperType(ts=-133.99)
        self.assertEqual(s.ts, "-133.99")

    def test_tostr_transforms_bool_into_str(self):
        s = SuperType(ts=False)
        self.assertEqual(s.ts, "False")



from __future__ import annotations
from jsonclasses import jsonclass


def increase_updated_count(book: CBMBook) -> None:
    book.updated_count += 1


@jsonclass(on_update=[increase_updated_count, increase_updated_count])
class CBMBook:
    name: str
    content: str
    updated_count: int = 0



from __future__ import annotations
from typing import Any, Optional
from jsonclasses import jsonclass, types

@jsonclass
class SimpleEq:
    eq_value: Optional[Any] = types.any.eq("dsadsa")

    ceq_value: Optional[Any] = types.any.eq(lambda: "dsadsa")
    teq_value: Optional[Any] = types.any.eq(types.default("dsadsa"))

@jsonclass
class SimpleNeq:
    neq_value: Optional[Any] = types.any.neq("dsadsa")

    cneq_value: Optional[Any] = types.any.neq(lambda: "dsadsa")
    tneq_value: Optional[Any] = types.any.neq(types.default("dsadsa"))



from __future__ import annotations
from unittest import TestCase
from tests.classes.cellphone_title import CellphoneTitle


class TestToTitle(TestCase):

    def test_totitle_titlizes_srt(self):
        product = CellphoneTitle(cellphone_name='hello', cellphone_title='sale for today')
        self.assertEqual(product.cellphone_title, 'Sale For Today')

    def test_totitle_titlizes_srt_with_special_characters(self):
        product = CellphoneTitle(cellphone_name='hello', cellphone_title='#sale !!$for t*^oday')
        self.assertEqual(product.cellphone_title, '#Sale !!$For T*^Oday')

    def test_totitle_titlizes_srt_with_int(self):
        product = CellphoneTitle(cellphone_name='hello', cellphone_title='123sale 2323for t77oday')
        self.assertEqual(product.cellphone_title, '123Sale 2323For T77Oday')


from jsonclasses import jsonclass, types

val = 0
intval = 0


def check_value() -> int:
    return val


def check_args() -> int:
    return intval


def callback(new_value: int) -> None:
    global val, intval
    val = val + 1
    intval = new_value


def callbackz() -> None:
    global val
    val = val + 10


@jsonclass
class UserOnsave:
    name: str = types.str.required
    age: int = types.int.onsave(callback).required


@jsonclass
class UserOnsaveZ:
    name: str = types.str.required
    age: int = types.int.onsave(callbackz).required



"""module for inputname modifier."""
from ..fdef import FDef, EnumInput
from .modifier import Modifier


class InputNameModifier(Modifier):
    """Input name modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        if fdef.enum_input is None:
            fdef._enum_input = EnumInput.NAME
        else:
            fdef._enum_input = fdef.enum_input | EnumInput.NAME
        return



"""module for assigning operator directly modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
from ..fdef import FDef
if TYPE_CHECKING:
    from ..ctx import Ctx


class AsopdModifier(Modifier):
    """Assigning operator directly modifier assigns the operator to the
    current field directly.
    """

    def define(self, fdef: FDef) -> None:
        fdef._requires_operator_assign = True

    def validate(self, ctx: Ctx) -> None:
        if ctx.holder.is_new or ctx.keypathr[-1] in ctx.holder.modified_fields:
            field = ctx.holder.__class__.cdef.field_named(ctx.keypathr[-1])
            fidname = ctx.holder.__class__.cdef.jconf.ref_name_strategy(field)
            if getattr(ctx.holder, fidname) is None:
                ctx.raise_vexc('no operator being assigned')



# from time import time
# from unittest import TestCase

# @classmethod
# def setUpClass(cls):
#     cls.startTime = time()


# @classmethod
# def tearDownClass(cls):
#     print("\n%s.%s: %.3f" % (cls.__module__, cls.__name__, time() - cls.startTime))


# TestCase.setUpClass = setUpClass
# TestCase.tearDownClass = tearDownClass



from __future__ import annotations
from typing import Optional, Tuple
from jsonclasses import jsonclass, types
from jsonclasses.ctx import Ctx

val = 0
oldval = 0
newval = 0


def check_value() -> int:
    return val


def check_args() -> Tuple[int, int]:
    return (oldval, newval)


def callback(old: int, new: int) -> None:
    global val, oldval, newval
    val = val + 1
    oldval = old
    newval = new


def callbackb(old: int, new: int) -> None:
    global val, oldval, newval
    val = val + 1
    oldval = old
    newval = new
    return new == old + 1


def callbacks(old: int, new: int, context: VCtx) -> None:
    global val, oldval, newval
    val = val + 1
    oldval = old
    newval = new
    if context.val == new + old:
        return None
    else:
        return 'invalid'


@jsonclass
class CompareUser:
    age: int = types.int.compare(callback).required
    name: Optional[str]


@jsonclass
class CompareUserB:
    age: int = types.int.compare(callbackb).required
    name: Optional[str]


@jsonclass
class CompareUserS:
    age: int = types.int.compare(callbacks).required
    name: Optional[str]



from __future__ import annotations
from unittest import TestCase
from jsonclasses import jsonclass
from jsonclasses.isjsonclass import isjsonclass
from jsonclasses.jconf import JConf
from jsonclasses.keypath import (
    camelize_key, identical_key, reference_key, underscore_key
)
from jsonclasses.cgraph import CGraph
from jsonclasses.excs import JSONClassRedefinitionException
from tests.classes.simple_account import SimpleAccount
from tests.classes.simple_company import SimpleCompany
from tests.classes.simple_employee import SimpleEmployee
from tests.funcs.yet_another_key_transformer import yet_another_key_transformer


class TestJsonClass(TestCase):

    def test_jsonclass_returns_jsonclass_without_arguments(self):
        self.assertTrue(isjsonclass(SimpleAccount))

    def test_jsonclass_returns_jsonclass_with_arguments(self):
        self.assertTrue(isjsonclass(SimpleCompany))

    def test_jsonclass_install_default_config_without_arguments(self):
        class_config = SimpleAccount.cdef.jconf
        default_config = JConf(cgraph='default',
                                output_key_strategy=camelize_key,
                                input_key_strategy=underscore_key,
                                strict_input=True,
                                ref_name_strategy=reference_key,
                                validate_all_fields=False,
                                abstract=False,
                                reset_all_fields=False,
                                output_null=False,
                                on_create=[],
                                on_update=[],
                                on_delete=[],
                                can_create=[],
                                can_update=[],
                                can_delete=[],
                                can_read=[])
        self.assertEqual(class_config, default_config)

    def test_jsonclass_cgraph_changes_cgraph(self):
        cgraph = SimpleCompany.cdef.jconf.cgraph
        company_graph = CGraph('simplecompany')
        self.assertEqual(cgraph, company_graph)

    def test_jsonclass_output_key_strategy_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.output_key_strategy, identical_key)

    def test_jsonclass_input_key_strategy_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.input_key_strategy, identical_key)

    def test_jsonclass_strict_input_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.strict_input, False)

    def test_jsonclass_key_transformer_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.ref_name_strategy,
            yet_another_key_transformer)

    def test_jsonclass_validate_all_fields_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.validate_all_fields, True)

    def test_jsonclass_abstract_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.abstract, True)

    def test_jsonclass_reset_all_fields_changes_config(self):
        self.assertEqual(
            SimpleEmployee.cdef.jconf.reset_all_fields, True)

    def test_jsonclass_raises_value_error_if_decorated_is_not_class(self):
        with self.assertRaisesRegex(ValueError, '@jsonclass should be used to'
                                                ' decorate a class\\.'):
            @jsonclass
            def _my_method():
                pass

    def test_jsonclass_raises_if_duplicate_names_on_same_graph(self):
        with self.assertRaisesRegex(JSONClassRedefinitionException,
                                    'jsonclass name conflict in graph'):
            @jsonclass(class_graph='simplecompany')
            class SimpleCompany:
                str_field: str
                int_field: str



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_str import SuperStr
from bcrypt import checkpw


class TestSalt(TestCase):

    def test_salt_add_salt_to_a_string(self):
        ss = SuperStr(password='123456')
        self.assertNotEqual(ss.password, '123456')
        self.assertTrue(checkpw('123456'.encode(), ss.password.encode()))



from __future__ import annotations
from unittest import TestCase
from tests.classes.onwrite import (OnwriteSingle, OnwriteDouble, OnwriteTriple,
                                   check_value)


class TestOnWrite(TestCase):

    def test_new_object_triggers_onwrite(self):
        val = check_value()
        item = OnwriteSingle(a='a', b=1)
        item.a = 'b'
        item._set_on_save()
        self.assertEqual(val + 1, check_value())

    def test_exist_object_wont_trigger_if_this_field_is_not_modified(self):
        val = check_value()
        item = OnwriteSingle(a='a', b=1)
        setattr(item, '_is_new', False)
        item.a = 'b'
        item._set_on_save()
        self.assertEqual(val, check_value())

    def test_exist_object_triggers_if_this_field_is_modified(self):
        val = check_value()
        item = OnwriteSingle(a='a', b=1)
        setattr(item, '_is_new', False)
        item.b = 5
        item._set_on_save()
        self.assertEqual(val + 1, check_value())

    def test_onwrite_accepts_value(self):
        val = check_value()
        item = OnwriteDouble(a='a', b=1)
        setattr(item, '_is_new', False)
        item.b = 5
        item._set_on_save()
        self.assertEqual(val + 5, check_value())

    def test_onwrite_accepts_value_and_context(self):
        val = check_value()
        item = OnwriteTriple(a='a', b=1)
        setattr(item, '_is_new', False)
        item.b = 5
        item._set_on_save()
        self.assertEqual(val + 10, check_value())



"""This modules contains the JSONClasses types modifier."""
from __future__ import annotations
from typing import Callable, Any, Optional
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from copy import deepcopy
from .jobject import JObject
from .fdef import FDef
from .modifiers import (BoolModifier, ChainedModifier, FValModifier, FObjModifier,
                        CompareModifier, DateModifier, DatetimeModifier,
                        DefaultModifier, DictOfModifier, EagerModifier,
                        EmbeddedModifier, EnumModifier, EqModifier, NeqModifier,
                        FloatModifier, IndexModifier, InputAllModifier, HexColorModifier,
                        InputNameModifier, InputLnameModifier, AtModifier,
                        InputValueModifier, InstanceOfModifier, ThisModifier,
                        IntModifier, InvalidModifier, LengthModifier,
                        LinkedByModifier, LinkedInModifier, GetterModifier,
                        LinkedThruModifier, LinkToModifier, ListOfModifier,
                        MatchModifier, MaxModifier, MaxlengthModifier,
                        LtModifier, MinModifier, MinlengthModifier, GtModifier,
                        NegativeModifier, NonnullModifier, NullableModifier,
                        OneOfModifier, UnionModifier, OnWriteModifier,
                        OnSaveModifier, OnUpdateModifier, OutputLnameModifier,
                        OutputNameModifier, OutputValueModifier, IsObjOfModifier,
                        PositiveModifier, PresentModifier, SetterModifier,
                        NonnegativeModifier, NonpositiveModifier,
                        PresentWithModifier, PresentWithoutModifier,
                        PreserializeModifier, PrimaryModifier, IsThisModifier,
                        RangeModifier, ReadonlyModifier, ReadwriteModifier,
                        RefereeModifier, ReferrerModifier, OneIsValidModifier,
                        RequiredModifier, ResetModifier, SetOnSaveModifier,
                        FSetOnSaveModifier, AddModifier, SubModifier,
                        DivModifier, ModModifier, MulModifier, GetOpModifier,
                        StrModifier, StrictModifier, AssignModifier,
                        TempModifier, TransformModifier, TrimModifier,
                        TruncateModifier, ValidateModifier, MongoIdModifier,
                        Modifier, WriteNonnullModifier, WriteonceModifier,
                        WriteonlyModifier, DenyModifier, CascadeModifier,
                        NullifyModifier, AsopModifier, FmtModifier,
                        AsopdModifier, UrlModifier, EmailModifier,
                        DigitModifier,AlphaModifier,NumericModifier,
                        AlnumModifier,ToTitleModifier,ToCapModifier,
                        ToLowerModifier,ToUpperModifier,RoundModifier,
                        UnresolvedModifier, AnyModifier, CeilModifier,
                        FloorModifier, BeforeModifier, AfterModifier,
                        ReverseModifier, ReplaceModifier, ReplacerModifier,
                        OddModifier, EvenModifier, AbsModifier,SplitModifier,
                        JoinModifier, AuthIdentityModifier, SaltModifier,
                        AuthByModifier, PassinModifier, CheckpwModifier,
                        ToListModifier, ToBoolModifier, ToFloatModifier,
                        ToIntModifier, ToStrModifier, RandomDigitsModifier,
                        RandomAlnumpuncsModifier, RandomAlnumsModifier,
                        RandomIntModifier, RandomFloatModifier,
                        ToBoSecModifier, ToBoMinModifier, ToBoHourModifier,
                        ToNextSecModifier, ToNextMinModifier, UploaderModifier,
                        ToNextHourModifier, CrossFetchModifier,
                        ToNextMonModifier, ToNextYearModifier, IsObjModifier,
                        ToNextDayModifier, ToBoYearModifier, ToBoMonModifier,
                        ToBoDayModifier, PadStartModifier, PadEndModifier,
                        UnqueryableModifier, QueryableModifier, PowModifier,
                        FormatDatetimeModifier, NoCopyModifier, SqrtModifier,
                        CanCModifier, CanUModifier, CanRModifier, MapModifier,
                        FilterModifier, HasPrefixModifier, HasSuffixModifier,
                        InsertAtModifier, AppendModifier, PrependModifier,
                        WrapIntoListModifier, IsPrefixOfModifier,
                        IsSuffixOfModifier, InverseModifier, UpperbondModifier,
                        LowerbondModifier, LenModifier, SecurepwModifier)

Str = str
Int = int
Float = float
Date = date
Datetime = datetime


class Types:
    """The class of types marks object. Types marks provide necessary
    information about an json object's shape, transformation, validation,
    serialization and sanitization.
    """

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self,
        original: Optional[Types] = None,
        *args: Modifier
    ) -> None:
        if not original:
            self.fdef = FDef()
            self.modifier = ChainedModifier()
        else:
            self.fdef = deepcopy(original.fdef)
            modifier = original.modifier
            for arg in args:
                modifier = modifier.append(arg)
                arg.define(self.fdef)
            self.modifier = modifier

    @property
    def invalid(self) -> Types:
        """Fields marked with invalid will never be valid, thus these fields
        will never pass validation.
        """
        return Types(self, InvalidModifier())

    @property
    def primary(self) -> Types:
        """Field marked with primary become the object's primary key.
        """
        return Types(self, ResetModifier(), PrimaryModifier())

    @property
    def readonly(self) -> Types:
        """Fields marked with readonly will not be able to go through
        initialization and set method. You can update value of these fields
        directly or through update method. This prevents client side to post
        data directly into these fields.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, ReadonlyModifier())

    @property
    def writeonly(self) -> Types:
        """Fields marked with writeonly will not be available in outgoing json
        form. Users' password is a great example of writeonly.
        """
        return Types(self, WriteonlyModifier())

    @property
    def readwrite(self) -> Types:
        """Fields marked with readwrite will be presented in both inputs and
        outputs. This is the default behavior. And this specifier can be
        omitted.
        """
        return Types(self, ReadwriteModifier())

    @property
    def writeonce(self) -> Types:
        """Fields marked with writeonce can only be set once through
        initialization and set method. You can update value of these fields
        directly or through update method. This is suitable for e.g. dating app
        user gender. Gender should not be changed once set.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, WriteonceModifier())

    @property
    def writenonnull(self) -> Types:
        """Fields marked with writenonnull can only be set to a nonnull value.
        The update method doesn't have this limitation. This prevents user from
        setting present value back into None.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, WriteNonnullModifier())

    @property
    def internal(self) -> Types:
        """Fields marked with internal will not be accepted as input, and it
        will not be present in output. These fields are internal and hidden
        from users.
        """
        return Types(self, ReadonlyModifier(), WriteonlyModifier())

    @property
    def temp(self) -> Types:
        """Fields marked with temp won't be written into database. As soon as
        database writing happens, value of this field is cleared and set to
        None. Examples of it's use cases are authentication code validation,
        input validation, etc.
        """
        return Types(self, TempModifier())

    @property
    def index(self) -> Types:
        """Fields marked with index are picked up by ORM integrations to setup
        database column index for you. This modifier doesn't have any effect
        around transforming and validating.
        """
        return Types(self, IndexModifier(False))

    def cindex(self, index_name: str) -> Types:
        """Fields marked with cindex have compound indexes. This modifier
        doesn't have any effect around transforming and validating.
        """
        return Types(self, IndexModifier(False, index_name))

    @property
    def unique(self) -> Types:
        """Fields marked with unique are picked up by ORM integrations to setup
        database column unique index for you. This modifier doesn't have any
        effect around transforming and validating. When database engine raises
        an exception, jsonclasses's web framework integration will catch it and
        return 400 automatically.

        If you are implementing jsonclasses ORM integration, you should use
        UniqueFieldException provided by jsonclasses.excs to keep
        consistency with other jsonclasses integrations.
        """
        return Types(self, IndexModifier(True))

    def cunique(self, index_name: str) -> Types:
        """Fields marked with cunique have compound indexes. This is a unique
        index. This modifier doesn't have any effect around transforming and
        validating.
        """
        return Types(self, IndexModifier(True, index_name))

    @property
    def queryable(self) -> Types:
        """Fields marked with queryable is queryable. This is the default
        behavior.
        """
        return Types(self, QueryableModifier())

    @property
    def unqueryable(self) -> Types:
        """Fields marked with unqueryable is not queryable.
        """
        return Types(self, UnqueryableModifier())

    @property
    def embedded(self) -> Types:
        """Instance fields marked with the embedded mark is embedded into the
        hosting document for noSQL databases.
        """
        return Types(self, EmbeddedModifier())

    def getter(self, calc: Callable | Types) -> Types:
        """Getter modifier marks a field as calculated field. It's not stored.
        """
        return Types(self, GetterModifier(calc))

    def setter(self, setter: Callable | Types) -> Types:
        """Setter modifier provides setter to calculated fields.
        """
        return Types(self, SetterModifier(setter))

    @property
    def nocopy(self) -> Types:
        """NoCopy modifier marks fields is nocopy.
        """
        return Types(self, NoCopyModifier())

    @property
    def linkto(self) -> Types:
        """In a database relationship, fields marked with linkto save an id of
        the object being referenced at the local table.
        """
        return Types(self, LinkToModifier())

    def linkedby(self, foreign_key: str) -> Types:
        """In a database relationship, fields marked with linkedby find
        reference from the destination table.
        """
        return Types(self, LinkedByModifier(foreign_key))

    def linkedthru(self, foreign_key: str) -> Types:
        """In a database relationship, fields marked with linkedthru save
        relationships to a designated association table and find references
        through it.
        """
        return Types(self, LinkedThruModifier(foreign_key))

    def linkedin(self, cls: Any) -> Types:
        """In a database relationship, fields marked with linkedin save
        relationships to the table under provided class.
        """
        return Types(self, LinkedInModifier(cls))

    def referrer(self, referrer_key: str) -> Types:
        """In a many to many database relationship, fields marked with referrer
        has a provided custom key name in the association table.
        """
        return Types(self, ReferrerModifier(referrer_key))

    def referee(self, referee_key: str) -> Types:
        """In a many to many database relationship, fields marked with referee
        reference the other side of the relationship with this provided custom
        key name.
        """
        return Types(self, RefereeModifier(referee_key))

    @property
    def nullify(self) -> Types:
        """When an object is deleted, linked objects' references are set to
        null instead of deleted.
        """
        return Types(self, NullifyModifier())

    @property
    def cascade(self) -> Types:
        """When an object is deleted, linked objects with cascade relationship
        are deleted.
        """
        return Types(self, CascadeModifier())

    @property
    def deny(self) -> Types:
        """When an object is deleted, linked objects with deny relationship
        prevent this object being deleted.
        """
        return Types(self, DenyModifier())

    @property
    def str(self) -> Types:
        """Fields marked with str should be str type. This is a type modifier.
        """
        return Types(self, StrModifier())

    def match(self, pattern: Str | Callable | Types) -> Types:
        """Fields marked with match are tested againest the argument regular
        expression pattern.
        """
        return Types(self, MatchModifier(pattern))

    def oneof(self, str_list: list[Str]) -> Types:
        """This is the enum equivalent for jsonclasses. Values in the provided
        list are considered valid values.
        """
        return Types(self, OneOfModifier(str_list))

    def minlength(self, length: int | Callable | Types) -> Types:
        """The value of iterable should be longer than the provided minlength.

        Args:
          length (int): The minimum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, MinlengthModifier(length))

    def maxlength(self, length: int | Callable | Types) -> Types:
        """The value of iterable should be shorter than the provided maxlength.

        Args:
          length (int): The maximum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, MaxlengthModifier(length))

    def length(self, minlength: int, maxlength: Optional[int] = None) -> Types:
        """Fields marked with length should have a length which is between the
        two arguments. If only one argument is provided, the value length
        should be exactly that length.

        Args:
          minlength (int): The minimum length required for the value.
          maxlength (Optional[int]): The maximum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, LengthModifier(minlength, maxlength))

    def add(self, by: int | float | Callable | Types) -> Types:
        """This modifier adds int or float value to original value

        """
        return Types(self, EagerModifier(), AddModifier(by))

    def sub(self, by: int | float | Callable | Types) -> Types:
        """This modifier for Int or float value subs original value
        """
        return Types(self, EagerModifier(), SubModifier(by))

    def mul(self, by: int | float | Callable | Types) -> Types:
        """This modifier for Int or float value muls by original value
        """
        return Types(self, EagerModifier(), MulModifier(by))

    def pow(self, by: int | float | Callable | Types) -> Types:
        """This modifer is used to transform the value of the int or float value
        to the power of the original value
        """
        return Types(self, EagerModifier(), PowModifier(by))

    def div(self, by: int | float | Callable | Types) -> Types:
        """This modifier for Int or float value divs by original value
        """
        return Types(self, EagerModifier(), DivModifier(by))

    def mod(self, by: int | float | Callable | Types) -> Types:
        """This modifier for Int or float value mods original value
        """
        return Types(self, EagerModifier(), ModModifier(by))

    def map(self, validate_callable: Callable) -> Types:
        """This modifier is used to getting a map object of the results
        """
        return Types(self, EagerModifier(), MapModifier(validate_callable))

    def filter(self, validate_callable: Callable) -> Types:
        """This modifier is used to filter the given sequence
        """
        return Types(self, EagerModifier(), FilterModifier(validate_callable))

    def hasprefix(self, validate_prefix: str | list[str | int] | Callable | Types) -> Types:
        """Hasprefix modifier validates if a string is prefix of another string
        or a list is prefix of another list
        """
        return Types(self, HasPrefixModifier(validate_prefix))

    def hassuffix(self, validate_suffix: str | list[str | int] | Callable | Types) -> Types:
        """Hassuffix modifier validates if a string is suffix of another string
        or a list is suffix of another list
        """
        return Types(self, HasSuffixModifier(validate_suffix))

    def isprefixof(self, prefix: str | list[str | int] | Callable | Types) -> Types:
        """Hassuffix modifier validates if a string is suffix of another string
        or a list is suffix of another list
        """
        return Types(self, IsPrefixOfModifier(prefix))

    def issuffixof(self, suffix: str | list[str | int] | Callable | Types) -> Types:
        """Hassuffix modifier validates if a string is suffix of another string
        or a list is suffix of another list
        """
        return Types(self, IsSuffixOfModifier(suffix))

    def insertat(self, item: Any | Callable | Types, index: int | Callable | Types) -> Types:
        """
        """
        return Types(self, EagerModifier(), InsertAtModifier(item, index))

    def prepend(self, item: str | int | float | Callable | Types) -> Types:
        """
        """
        return Types(self, EagerModifier(), PrependModifier(item))

    def append(self, item: str | int | float | Callable | Types) -> Types:
        """
        """
        return Types(self, EagerModifier(), AppendModifier(item))

    @property
    def len(self) -> Types:
        """Return the length of value.
        """
        return Types(self, EagerModifier(), LenModifier())

    def upperbond(self, max_value: int | float | Callable | Types) -> Types:
        """Decrease the value of the field to upperbond if the value of the
        field is larger than upperbond.
        """
        return Types(self, EagerModifier(), UpperbondModifier(max_value))

    def lowerbond(self, min_value: int | float | Callable | Types) -> Types:
        """Increase the value of the field to lowerbond if the value of the
        field is smaller than lowerbond.
        """
        return Types(self, EagerModifier(), LowerbondModifier(min_value))

    @property
    def inverse(self) -> Types:
        """Change the value to false if value is true, vice versa.
        """
        return Types(self, EagerModifier(), InverseModifier())

    @property
    def wrapintolist(self) -> Types:
        """
        """
        return Types(self, WrapIntoListModifier())

    @property
    def sqrt(self) -> Types:
        """This modifier is used to transform the square root of any number.
        """
        return Types(self, EagerModifier(), SqrtModifier())

    @property
    def url(self) -> Types:
        """Fields marked with url should be valid url string.
        """
        return Types(self, UrlModifier())

    @property
    def digit(self) -> Types:
        """Values of fields marked with digit should be valid digit string.
        """
        return Types(self, DigitModifier())

    @property
    def alpha(self) -> Types:
        """Values of fields marked with alpha should be valid alpha string.
        """
        return Types(self, AlphaModifier())

    @property
    def numeric(self) -> Types:
        """Values of fields marked with numeric should be valid numeric string.
        """
        return Types(self, NumericModifier())

    @property
    def email(self) -> Types:
        """Values of fields marked with email should be valid email format.
        """
        return Types(self, EmailModifier())

    @property
    def hexcolor(self) -> Types:
        """Values of fields marked with hexcolor should be valid hexcolor format.
        """
        return Types(self, HexColorModifier())

    @property
    def securepw(self) -> Types:
        """Values of fields marked with password should be valid secure password
        format.
        """
        return Types(self, SecurepwModifier())

    @property
    def alnum(self) -> Types:
        """Values fields marked with alnum should be valid alnum strings.
        """
        return Types(self, AlnumModifier())

    @property
    def int(self) -> Types:
        """Fields marked with int should be int type. This is a type modifier.
        """
        return Types(self, IntModifier())

    @property
    def odd(self) -> Types:
        """Fields marked with int should be odd. This is a int type modifier.
        """
        return Types(self, OddModifier())

    @property
    def even(self) -> Types:
        """Fields marked with int should be even. This is a int type modifier.
        """
        return Types(self, EvenModifier())

    @property
    def float(self) -> Types:
        """Fields marked with float should be float type. This is a type
        modifier.
        """
        return Types(self, FloatModifier())

    def min(self, value: Float | Callable | Types) -> Types:
        """Fields marked with min are tested again this value. Values less than
        the argument value are considered invalid.
        """
        return Types(self, MinModifier(value))

    def max(self, value: Float | Callable | Types) -> Types:
        """Fields marked with max are tested again this value. Values greater
        than the argument value are considered invalid.
        """
        return Types(self, MaxModifier(value))

    def lte(self, value: Float | Callable | Types) -> Types:
        """Fields marked with lte are tested again this value. Values greater
        than the argument value are considered invalid.
        """
        return Types(self, MaxModifier(value))

    def gte(self, value: Float | Callable | Types) -> Types:
        """Fields marked with gte are tested again this value. Values less than
        the argument value are considered invalid.
        """
        return Types(self, MinModifier(value))

    def lt(self, value: Float | Callable | Types) -> Types:
        """Fields marked with lt are tested again this value. Values greater
        or equal than the argument value are considered invalid.
        """
        return Types(self, LtModifier(value))

    def gt(self, value: Float | Callable | Types) -> Types:
        """Fields marked with gt are tested again this value. Values less than
        or equal than the argument value are considered invalid.
        """
        return Types(self, GtModifier(value))


    def range(self, min_value: Float | Callable | Types, max_value: Float | Callable | Types) -> Types:
        """Fields marked with range are tested again argument values. Only
        values between the arguments range are considered valid.
        """
        return Types(self, RangeModifier(min_value, max_value))

    @property
    def negative(self) -> Types:
        """Fields marked with negative should have a value less than zero.
        """
        return Types(self, NegativeModifier())

    @property
    def positive(self) -> Types:
        """Fields marked with negative should have a value greater than zero.
        """
        return Types(self, PositiveModifier())

    @property
    def nonnegative(self) -> Types:
        """Fields marked with nonnegative should have a value greater than or
        equal to zero.
        """
        return Types(self, NonnegativeModifier())

    @property
    def nonpositive(self) -> Types:
        """Fields marked with nonnegative should have a value less than or
        equal to zero.
        """
        return Types(self, NonpositiveModifier())

    @property
    def bool(self) -> Types:
        """Fields marked with bool should be bool type. This is a type modifier.
        """
        return Types(self, BoolModifier())

    @property
    def date(self) -> Types:
        """Fields marked with date should be date type. This is a type modifier.
        """
        return Types(self, DateModifier())

    @property
    def datetime(self) -> Types:
        """Fields marked with datetime should be datetime type. This is a type
        modifier.
        """
        return Types(self, DatetimeModifier())

    def before(self, point: Date | Datetime | Callable | Types) -> Types:
        """InputDate should be before the date. This is a date modifier
        """
        return Types(self, BeforeModifier(point))

    def after(self, point: Date | Datetime | Callable | Types) -> Types:
        """InputDate should be after the date. This is a date modifier
        """
        return Types(self, AfterModifier(point))

    @property
    def tobosec(self) -> Types:
        """This modifier empty the microsecond in datetime.
        """
        return Types(self, EagerModifier(), ToBoSecModifier())

    @property
    def tobomin(self) -> Types:
        """This modifier empty the microsecond and second in datetime
        """
        return Types(self, EagerModifier(), ToBoMinModifier())

    @property
    def tobohour(self) -> Types:
        """This modifier empty microsecond, second and minute in datetime
        """
        return Types(self, EagerModifier(), ToBoHourModifier())

    @property
    def tonextsec(self) -> Types:
        """This modifier Go to the next second in datetime
        """
        return Types(self, EagerModifier(), ToNextSecModifier())

    @property
    def tonextmin(self) -> Types:
        """This modifier Go to the next minute in datetime
        """
        return Types(self, EagerModifier(), ToNextMinModifier())

    @property
    def tonexthour(self) -> Types:
        """This modifier Go to the next hour in datetime
        """
        return Types(self, EagerModifier(), ToNextHourModifier())

    @property
    def tonextyear(self) -> Types:
        """This modifier Go to the next year in datetime or in date
        """
        return Types(self, EagerModifier(), ToNextYearModifier())

    @property
    def tonextmon(self) -> Types:
        """This modifier Go to the next month in datetime or in date
        """
        return Types(self, EagerModifier(), ToNextMonModifier())

    @property
    def tonextday(self) -> Types:
        """This modifier Go to the next day in datetime or in date
        """
        return Types(self, EagerModifier(), ToNextDayModifier())

    @property
    def toboyear(self) -> Types:
        """This modifier empty microsecond, second, minute, hour, day and month
        in datetime or empty day, month in date
        """
        return Types(self, EagerModifier(), ToBoYearModifier())

    @property
    def tobomon(self) -> Types:
        """This modifier empty microsecond, second, minute, hour and day in
        datetime or empty day in date
        """
        return Types(self, EagerModifier(), ToBoMonModifier())

    @property
    def toboday(self) -> Types:
        """This modifier empty microsecond, second, minute and hour in
        datetime or return date directly
        """
        return Types(self, EagerModifier(), ToBoDayModifier())

    def fmtd(self, format: str | Callable | Types) -> Types:
        """This modifier format datetime or date value.
        """
        return Types(self, FormatDatetimeModifier(format))

    def enum(self, enum_class: type[Enum] | str) -> Types:
        """Fields marked with enum should be enum value of provided enum type.
        This is a type modifier.
        """
        return Types(self, EnumModifier(enum_class))

    @property
    def inputall(self) -> Types:
        """Inputall makes enum field to accept all kinds of acceptable enum
        values in any forms.
        """
        return Types(self, InputAllModifier())

    @property
    def inputlname(self) -> Types:
        """Inputlname makes enum field to accept enum's lowercase name as
        input.
        """
        return Types(self, InputLnameModifier())

    @property
    def inputname(self) -> Types:
        """Inputlname makes enum field to accept enum's uppercase name as
        input.
        """
        return Types(self, InputNameModifier())

    @property
    def inputvalue(self) -> Types:
        """Inputlname makes enum field to accept enum's value as input.
        """
        return Types(self, InputValueModifier())

    @property
    def outputlname(self) -> Types:
        """Outputlname makes enum field to output lowercase name as display
        value.
        """
        return Types(self, OutputLnameModifier())

    @property
    def outputname(self) -> Types:
        """Outputname makes enum field to output uppercase name as display
        value.
        """
        return Types(self, OutputNameModifier())

    @property
    def outputvalue(self) -> Types:
        """Outputvalue makes enum field to output value as display value.
        """
        return Types(self, OutputValueModifier())

    def listof(self, item_types: Any) -> Types:
        """Fields marked with listof should be a list of the given type. This
        is a type modifier.
        """
        return Types(self, ListOfModifier(item_types))

    def dictof(self, item_types: Any) -> Types:
        """Fields marked with listof should be a str keyed dict of the given
        type. This is a type modifier.
        """
        return Types(self, DictOfModifier(item_types))

    @property
    def strict(self) -> Types:
        """Object fields marked with strict disallow undefined keys.
        """
        return Types(self, StrictModifier())

    def objof(self, jcls: Any) -> Types:
        """Fields marked with objof are objects of given class.
        """
        return Types(self, InstanceOfModifier(jcls))

    def union(self, type_list: list[Any]) -> Types:
        """Fields marked with union accepts value from these types.
        """
        return Types(self, UnionModifier(type_list))

    @property
    def any(self) -> Types:
        """Fields marked with any can be any value.
        """
        return Types(self, AnyModifier())

    @property
    def required(self) -> Types:
        """Fields marked with required are invalid when value is None.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, RequiredModifier())

    @property
    def nullable(self) -> Types:
        """Fields marked with nullable can be None. This is the default
        behavior even without this modifier. It's the opposite to required
        modifier. Values inside lists have implicitly required modifier. Use
        this to allow null or None values inside lists.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, NullableModifier())

    @property
    def present(self) -> Types:
        """When validating, field marked with present, can not pass validation
        if it has a None value. This is useful for foreign key fields to do
        required validation.
        """
        return Types(self, PresentModifier())

    def presentwith(self, referring_key: str) -> Types:
        """Fields marked with presentwith modifier are forced presented if
        referring field is present. If referring field has None value, this
        field's value is optional. If referring field has non None value, value
        of this field is required.
        """
        return Types(self, PresentWithModifier(referring_key))

    def presentwithout(self, referring_keys: Str | list[Str]) -> Types:
        """Fields marked with presentwithout modifier are forced presented if
        referring field is not present. If referring field has None value, this
        field's value should be present. If referring field has non None value,
        value of this field is not forced to be present.
        """
        return Types(self, PresentWithoutModifier(referring_keys))

    def validate(self, validator: Callable | Types) -> Types:
        """The validate field mark takes a modifier callable as its sole
        argument. Use this to define custom field value validations.

        Args:
            validator (Callable): The validate callable takes up to 2
            arguments, which are value and context. Returning None or True
            means the value is valid, while returning a str message or False
            means validation failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, ValidateModifier(validator))

    def vmsg(self, validator: Callable | Types, message: str) -> Types:
        """The validate field mark takes a modifier callable as its sole
        argument. Use this to define custom field value validations.

        Args:
            validator (Callable | Types): The validate callable takes up to 2
            arguments, which are value and context. Returning None or True
            means the value is valid, while returning a str message or False
            means validation failed.
            msg (str): The error message to be outputted.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, ValidateModifier(validator, message))

    def compare(self, compare_callable: Callable) -> Types:
        """The compare field mark takes a modifier callable as its sole
        argument. If value of compared fields are changed, this modifier is
        called with 2 to 5 arguments.

        Args:
            compare_callable (Callable): The compare callable. Arg 1 is the old
            value, arg 2 is the new value, arg 3 is key path, arg 4 is the
            object, arg 5 is the context. Returning None means the value is
            valid, while returning a str message means validation failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self,
                     ResetModifier(),
                     CompareModifier(compare_callable))

    # transformers

    def default(self, value: Any) -> Types:
        """During initialization, if values of fields with default are not
        provided. The default value is used instead of leaving blank.

        Args:
          value (any): The default value of this field. If the value is
          callable, it's return value is used.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, DefaultModifier(value))

    def truncate(self, max_length: Int | Callable | Types) -> Types:
        """During initialization and set, if string value is too long, it's
        truncated to argument max length.

        Args:
          max_length (int): The allowed max length of the field value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TruncateModifier(max_length))

    @property
    def trim(self) -> Types:
        """This modifier will trim string value. Remove leading and trailing
        whitespaces.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TrimModifier())

    @property
    def totitle(self) -> Types:
        """This modifier titlizes string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToTitleModifier())

    @property
    def tocap(self) -> Types:
        """This modifier capitalizes string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToCapModifier())

    @property
    def tolower(self) -> Types:
        """This modifier lowercasefies string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToLowerModifier())

    @property
    def toupper(self) -> Types:
        """This modifier uppercasefies string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToUpperModifier())

    def padstart(self, char: str | Callable | Types, length: int | Callable | Types) -> Types:
        """This modifier padstart string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), PadStartModifier(char, length))

    def padend(self, char: str | Callable | Types, length: int | Callable | Types) -> Types:
        """This modifier padstart string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), PadEndModifier(char, length))

    @property
    def tolist(self) -> Types:
        """This modifier transforms value into list.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToListModifier())

    @property
    def tobool(self) -> Types:
        """This modifier transforms value into bool.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToBoolModifier())

    @property
    def tofloat(self) -> Types:
        """This modifier transforms value into float.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToFloatModifier())

    @property
    def toint(self) -> Types:
        """This modifier transforms value into int.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToIntModifier())

    @property
    def tostr(self) -> Types:
        """This modifier transforms value into str.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToStrModifier())

    def replace(self, old: Str | Callable | Types, new: Str | Callable | Types) -> Types:
        """Replace modifier replaces occurance with substitutions.
        """
        return Types(self, EagerModifier(), ReplaceModifier(old, new))

    def replacer(self, reg: Str | Callable | Types, rep: Str | Callable | Types) -> Types:
        """Replacer modifier replaces occurance matches regular expression with
        replacement string.
        """
        return Types(self, EagerModifier(), ReplacerModifier(reg, rep))

    def split(self, sep: Str | Callable | Types) -> Types:
        """Split modifier splits string into a list of strings.
        """
        return Types(self, EagerModifier(), SplitModifier(sep))

    def join(self, sep: Str | Callable | Types) -> Types:
        """Join modifier concatenates a list of strings into a single string.
        """
        return Types(self, EagerModifier(), JoinModifier(sep))

    @property
    def salt(self) -> Types:
        """Salt modifier add salt to a string.
        """
        return Types(self, EagerModifier(), SaltModifier())

    @property
    def round(self) -> Types:
        """This modifier rounds number value.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), RoundModifier())

    @property
    def ceil(self) -> Types:
        """This modifier ceil number value.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), CeilModifier())

    @property
    def floor(self) -> Types:
        """This modifier floor number value.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), FloorModifier())

    @property
    def abs(self) -> Types:
        """This modifier abs number value.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), AbsModifier())

    def transform(self, transformer: Callable | Types) -> Types:
        """This mark applies transfromer on the value. When value is None, the
        transformer is not called. This class barely means to transform. Use
        default mark with a callable to assign calculated default value.

        Args:
          transformer (Callable): This transformer function takes one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TransformModifier(transformer))

    def fmt(self, formatter: Callable | Types) -> Types:
        """Fmt formats value.

        Args:
            formatter (Callable | Types): Describes the formation.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, FmtModifier(formatter))

    def reverse(self) -> Types:
        """This modifier reverse iterable value

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ReverseModifier())

    def uploader(self, uploader: Callable | str) -> Types:
        """Uploader is barely a syntax alias for transformer. Uploaded files
        can be processed through a transformer and get url attached to this
        object.

        Args:
            uploader (Callable | str): An uploader name or a transformer
            callable.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), UploaderModifier(uploader))

    def asop(self, asop_transformer: Callable) -> Types:
        """Asop modifier assigns transformed operator value to this field. When
        the operator is not present, a ValidationException is raised.

        Args:
            asop_transformer (Callable): This transformer function takes 1 to
            3 arguments. The first one is the opeartor, the second one is the
            value of the field, the third is the transforming context.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, AsopModifier(asop_transformer))

    @property
    def asopd(self) -> Types:
        """Asopd modifier assigns operator value directly to this field without
        any transforming. When the operator  is not present, a
        ValidationException is raised.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, AsopdModifier())

    def setonsave(self, setter: Callable) -> Types:
        """Setonsave modifier marks a field to be updated just before
        serializing into the database if this field is modified and to be
        serialized.

        Args:
          setter (Callable): This setter function takes zero or one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, PreserializeModifier(), SetOnSaveModifier(setter))

    def fsetonsave(self, setter: Callable) -> Types:
        """Fsetonsave modifier marks a field to be updated just before
        serializing into the database regardless of this field is modified.

        Args:
          setter (Callable): This setter function takes zero or one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, PreserializeModifier(), FSetOnSaveModifier(setter))

    def onsave(self, callback: Callable) -> Types:
        """Onsave inserts a callback into the modifier chain. If save action
        is triggered, the callback is called with the current value.

        Args:
            callback (Callable): This callback function takes one argument
            which is the current value of the field.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, OnSaveModifier(callback))

    def onupdate(self, callback: Callable) -> Types:
        """Onupdate is a callback modifier. If value updated when saving, this
        callback is called with the current value or both values.

        Args:
            callback (Callable): A callable which takes arguments.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, ResetModifier(), OnUpdateModifier(callback))

    def onwrite(self, callback: Callable) -> Types:
        """Onwrite is a callback modifier. Whenever a new value is being
        serialized into the database, this is called.

        Args:
            callback (Callable): A callable which takes arguments.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, OnWriteModifier(callback))

    @property
    def nonnull(self) -> Types:
        """This modifier is a instructional transformer designated for dictof
        and listof. This is not a modifier. To mark a field is required and
        should not be null, use `required` instead. This transformer should
        be used right before dictof and listof, to given an instruction of not
        leaving null for the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, NonnullModifier())

    # compound

    @property
    def mongoid(self: Types) -> Types:
        """This modifier assigns a default bson object id to the field.
        """
        return Types(self, MongoIdModifier())

    @property
    def tscreated(self: Types) -> Types:
        """This modifier adds a default current time to the field.
        """
        return Types(self, DefaultModifier(lambda: datetime.now(timezone.utc)))

    @property
    def tsupdated(self: Types) -> Types:
        """This modifier adds a default current time and a set on save.
        """
        return Types(self,
                     DefaultModifier(datetime.now),
                     PreserializeModifier(),
                     SetOnSaveModifier(lambda: datetime.now(timezone.utc)))

    def umininterval(self: Types, interval: timedelta) -> Types:
        """This modifier compares old and new value against the time interval.
        """
        def compare_callable(old: datetime, new: datetime):
            if new >= old + interval:
                return None
            return 'time interval too short'
        return Types(self,
                     ResetModifier(),
                     CompareModifier(compare_callable))

    # authorization

    @property
    def authidentity(self: Types) -> Types:
        """Fields marked with authidentity are used for authorization.
        """
        return Types(self, AuthIdentityModifier())

    def authby(self: Types, checker: Types) -> Types:
        """Fields marked with authby are used for authorization.
        """
        return Types(self, AuthByModifier(checker))

    @property
    def authbycheckpw(self: Types) -> Types:
        """This is a shortcut to `authby(types.checkpw(types.passin))`.
        """
        return Types(self, AuthByModifier(types.checkpw(types.passin)))


    def canc(self, checker: Callable | Types) -> Types:
        """CanC modifier validates value against the operator object user
        passed in. This modifier is special and doesn't bypass None value. If
        the operator is not present, this modifier fails.

        Args:
            checker (Callable | Types): The checker takes 1 to 4 arguments. The
            first is the operator object, the second is the object being
            operated, the third is the value of the field, the fourth is the
            validating context. Returning None or True means the value is
            valid, while returning a str message or False means validation
            failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, CanCModifier(checker))

    def canu(self, checker: Callable | Types) -> Types:
        """CanU modifier validates value against the operator object user
        passed in. This modifier is special and doesn't bypass None value. If
        the operator is not present, this modifier fails.

        Args:
            checker (Callable | Types): The checker takes 1 to 4 arguments. The
            first is the operator object, the second is the object being
            operated, the third is the value of the field, the fourth is the
            validating context. Returning None or True means the value is
            valid, while returning a str message or False means validation
            failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, CanUModifier(checker))

    def canw(self, checker: Callable | Types) -> Types:
        """CanW modifier validates value against the operator object user
        passed in. This modifier is special and doesn't bypass None value. If
        the operator is not present, this modifier fails.

        Args:
            checker (Callable | Types): The checker takes 1 to 4 arguments. The
            first is the operator object, the second is the object being
            operated, the third is the value of the field, the fourth is the
            validating context. Returning None or True means the value is
            valid, while returning a str message or False means validation
            failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, CanCModifier(checker), CanUModifier(checker))

    def canr(self, checker: Callable | Types) -> Types:
        """CanR modifier validates value against the operator object user
        passed in. This modifier is special and doesn't bypass None value. If
        the operator is not present, this modifier fails.

        Args:
            checker (Callable | Types): The checker takes 1 to 4 arguments. The
            first is the operator object, the second is the object being
            operated, the third is the value of the field, the fourth is the
            validating context. Returning None or True means the value is
            valid, while returning a str message or False means validation
            failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, CanRModifier(checker))

    # calc pipeline

    @property
    def passin(self: Types) -> Types:
        """Pass in modifier users passin value as the result.
        """
        return Types(self, EagerModifier(), PassinModifier())

    def checkpw(self: Types, against: Types) -> Types:
        """Checkpw modifier uses bcrypt's checkpw function to validate str
        value.
        """
        return Types(self, CheckpwModifier(against))

    def randomdigits(self: Types, length: int | Callable | Types) -> Types:
        """Random digits modifier generates a random digits string of length.
        """
        return Types(self, RandomDigitsModifier(length))

    def randomalnums(self: Types, length: Int | Callable | Types) -> Types:
        """Random alnums modifier generates a random alnums string of length.
        """
        return Types(self, RandomAlnumsModifier(length))

    def randomalnumpuncs(self: Types, length: int | Callable | Types) -> Types:
        """Random alnumpuncs modifier generates a random alnumpuncs string of
        length.
        """
        return Types(self, RandomAlnumpuncsModifier(length))

    def randomint(self: Types, min_value: int | float | Callable | Types,
                    max_value: int | float | Callable | Types) -> Types:
        """Random int modifier generates a random int value."""
        return Types(self, RandomIntModifier(min_value, max_value))

    def randomfloat(self: Types, min_value: int | float | Callable | Types,
                    max_value: int | float | Callable | Types) -> Types:
        """Random float modifier generates a random float value."""
        return Types(self, EagerModifier(), RandomFloatModifier(min_value, max_value))

    def crossfetch(self, cn: str, sk: str, fk: Optional[str] = None) -> Types:
        """Fetch a class with value matches this object's value at key.
        """
        return Types(self, CrossFetchModifier(cn, sk, fk))

    def fval(self: Types, field_name: str | Callable | Types) -> Types:
        """Get value at field from a JSONClass object.
        """
        return Types(self, FValModifier(field_name))

    def fobj(self: Types, field_name: str | Callable | Types) -> Types:
        """Get referenced object at field from a JSONClass object. Include it if
        it's not exist.
        """
        return Types(self, FObjModifier(field_name))

    def eq(self: Types, val: Any | Callable | Types) -> Types:
        """Eq modifier validates value by equal testing.
        """
        return Types(self, EqModifier(val))

    def neq(self: Types, val: Any | Types | Callable) -> Types:
        """Neq modifier validates value by unequal testing.
        """
        return Types(self, NeqModifier(val))

    @property
    def this(self: Types) -> Types:
        """Get the owner object of this field.
        """
        return Types(self, ThisModifier())

    def at(self: Types, index: Any | Callable | Types) -> Types:
        """At modifier returns result with subscription index.
        """
        return Types(self, AtModifier(index))

    def assign(self: Types, name: str, value: Any | Types | Callable) -> Types:
        """Assign value to name of object.
        """
        return Types(self, AssignModifier(name, value))

    @property
    def isthis(self: Types) -> Types:
        """Check whether the current value is the owner object.
        """
        return Types(self, IsThisModifier())

    def oneisvalid(self: Types, subroutines: list[Callable | Types]) -> Types:
        """One is valid is valid if one of subroutines is valid.
        """
        return Types(self, OneIsValidModifier(subroutines))

    def isobj(self: Types, getter: Types) -> Types:
        """Valid if objects are equal.
        """
        return Types(self, IsObjModifier(getter))

    def isobjof(self: Types, cls: type[JObject] | str) -> Types:
        """Valid if the value is object of a class.
        """
        return Types(self, IsObjOfModifier(cls))

    @property
    def getop(self: Types) -> Types:
        """Get the operator of this action.
        """
        return Types(self, GetOpModifier())

    # internal

    def _unresolved(self: Types, arg: Str) -> Types:
        """This modifier marks unresolved status. This is used internally. Do
        not use this.
        """
        return Types(self, UnresolvedModifier(arg))


types = Types()
"""The root of the types modifier. To mark an field with type annotation,
accessor annotation, modifier annotation and transformer annotation, use types
like this:

  @jsonclass
  class MyObject:
    enabled: bool = types.bool.readonly.required
    password: str = types.str.writeonly.length(8, 16).salt.required
"""



"""
This module defines `CDef`. Each JSON class has its own class
definition. The class definition object contains detailed information about how
user defines a JSON class. This is used by the framework to lookup class fields
and class field settings.
"""
from __future__ import annotations
from jsonclasses.jobject import JObject
from typing import Optional, Any, final, cast, TYPE_CHECKING
from dataclasses import fields as dataclass_fields
from .jfield import JField
from .fdef import FStore, DeleteRule, FType
from .rtypes import rtypes, rnamedtypes
if TYPE_CHECKING:
    from .jconf import JConf


@final
class CDef:
    """Class definition represents the class definition of JSON classes. Each
    JSON class has its own class definition. The class definition object
    contains detailed information about how user defines a JSON class. This is
    used by the framework to lookup class fields and class field settings.
    """

    def __init__(self: CDef, cls: type[JObject], jconf: JConf) -> None:
        """
        Initialize a new class definition.

        Args:
            cls (type): The JSON class for which the class definition is \
                created.
            jconf (JConf): The configuration object for the targeted class.
        """
        from .types import Types
        self._ref_names_resolved = False
        self._ref_types_resolved = False
        self._cls = cls
        jconf._cls = cls
        self._name: str = cls.__name__
        self._jconf: JConf = jconf
        self._list_fields: list[JField] = []
        self._dict_fields: dict[str, JField] = {}
        self._primary_field: Optional[JField] = None
        self._calc_fields: list[JField] = []
        self._setter_fields: list[JField] = []
        self._deny_fields: list[JField] = []
        self._nullify_fields: list[JField] = []
        self._cascade_fields: list[JField] = []
        self._field_names: list[str] = []
        self._camelized_field_names: list[str] = []
        self._reference_names: list[str] = []
        self._camelized_reference_names: list[str] = []
        self._list_reference_names: list[str] = []
        self._camelized_list_reference_names: list[str] = []
        self._virtual_reference_names: list[str] = []
        self._camelized_virtual_reference_names: list[str] = []
        self._virtual_reference_fields: dict[str, JField] = {}
        self._unique_fields: list[JField] = []
        self._assign_operator_fields: list[JField] = []
        self._auth_identity_fields: list[JField] = []
        self._auth_by_fields: list[JField] = []
        self._rfmap: dict[str, JField] = {}
        for field in dataclass_fields(cls):
            name = field.name
            self._field_names.append(name)
            if isinstance(field.default, Types):
                types = field.default
                default = None
            elif field.default == cast(Any, field).default_factory:
                types = rtypes(field.type)
                default = None
            else:
                types = rtypes(field.type)
                default = field.default
            types.fdef._cdef = self
            jfield = JField(cdef=self, name=name, default=default, types=types)
            self._camelized_field_names.append(jfield.json_name)
            self._list_fields.append(jfield)
            self._dict_fields[name] = jfield
            if types.fdef._primary:
                self._primary_field = jfield
            if types.fdef._fstore == FStore.CALCULATED:
                self._calc_fields.append(jfield)
            if types.fdef._setter is not None:
                self._setter_fields.append(jfield)
            if types.fdef._delete_rule == DeleteRule.DENY:
                self._deny_fields.append(jfield)
            elif types.fdef._delete_rule == DeleteRule.NULLIFY:
                self._nullify_fields.append(jfield)
            elif types.fdef._delete_rule == DeleteRule.CASCADE:
                self._cascade_fields.append(jfield)
            if types.fdef._unique:
                self._unique_fields.append(jfield)
            if types.fdef._requires_operator_assign:
                self._assign_operator_fields.append(jfield)
            if types.fdef._auth_identity:
                self._auth_identity_fields.append(jfield)
            if types.fdef._auth_by:
                self._auth_by_fields.append(jfield)
        self._tuple_fields: tuple[JField, ...] = tuple(self._list_fields)

    @property
    def cls(self: CDef) -> type:
        """The JSON class on which this class definition is defined.
        """
        return self._cls

    @property
    def name(self: CDef) -> str:
        """The name of the JSON class on which this class definition is
        defined.
        """
        return self._name

    @property
    def jconf(self: CDef) -> JConf:
        """The configuration object of the JSON class on which this class
        definition is defined.
        """
        return self._jconf

    def field_named(self: CDef, name: str) -> JField:
        """
        Get the field which is named `name`.

        Args:
            name (str): The name of the field to return.

        Returns:
            JField: The field named `name`.

        Raises:
            ValueError: If can't find a field with name `name`.
        """
        if not self._dict_fields.get(name):
            raise ValueError(f'no field named {name} in class definition')
        return self._dict_fields[name]

    @property
    def fields(self: CDef) -> tuple[JField, ...]:
        """Get the fields of this class definition as a tuple. This is useful
        for looping and iterating.
        """
        return self._tuple_fields

    @property
    def calc_fields(self: CDef) -> list[JField]:
        """Calculated fields of this class definition.
        """
        return self._calc_fields

    @property
    def calc_field_names(self: CDef) -> list[str]:
        """Names of calculated fields.
        """
        if hasattr(self, '_calc_field_names'):
            return self._calc_field_names
        self._calc_field_names = list(map(lambda f: f.name, self._calc_fields))
        return self._calc_field_names

    @property
    def setter_fields(self: CDef) -> list[JField]:
        """Calculated fields with setter of this class definition.
        """
        return self._setter_fields

    @property
    def setter_field_names(self: CDef) -> list[str]:
        """Names of calculated fields with setter.
        """
        if hasattr(self, '_setter_field_names'):
            return self._setter_field_names
        self._setter_field_names = [f.name for f in self._setter_fields]
        return self._setter_field_names

    @property
    def deny_fields(self: CDef) -> list[JField]:
        """Reference fields with deny delete rule.
        """
        return self._deny_fields

    @property
    def nullify_fields(self: CDef) -> list[JField]:
        """Reference fields with nullify delete rule.
        """
        return self._nullify_fields

    @property
    def cascade_fields(self: CDef) -> list[JField]:
        """Reference fields with cascade delete rule.
        """
        return self._cascade_fields

    @property
    def primary_field(self: CDef) -> Optional[JField]:
        """The class definition's primary field. This can be None if it's not
        defined by user.
        """
        return self._primary_field

    @property
    def unique_fields(self: CDef) -> list[JField]:
        """The unique fields of this class definition.
        """
        return self._unique_fields

    @property
    def assign_operator_fields(self: CDef) -> list[JField]:
        """The class definition's fields which require operator assigning on
        object creation.
        """
        return self._assign_operator_fields

    @property
    def auth_identity_fields(self: CDef) -> list[JField]:
        """Auth identity fields.
        """
        return self._auth_identity_fields

    @property
    def auth_by_fields(self: CDef) -> list[JField]:
        """Auth by fields.
        """
        return self._auth_by_fields

    @property
    def available_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return self._available_names

    @property
    def update_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return self._update_names

    @property
    def reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._reference_names)

    @property
    def camelized_reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._camelized_reference_names)

    @property
    def list_reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._list_reference_names)

    @property
    def camelized_list_reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._camelized_list_reference_names)

    @property
    def virtual_reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._virtual_reference_names)

    @property
    def camelized_virtual_reference_names(self: CDef) -> set[str]:
        self._resolve_ref_names_if_needed()
        return set(self._camelized_virtual_reference_names)

    @property
    def virtual_reference_fields(self: CDef) -> dict[str, JField]:
        self._resolve_ref_names_if_needed()
        return self._virtual_reference_fields

    def rname_to_jfield(self: CDef, ref_name: str) -> JField:
        self._resolve_ref_names_if_needed()
        return self._rfmap[ref_name]

    def rfield(
            self: CDef, fcls: type[JObject], fname: Optional[str],
            fkey: Optional[str]) -> Optional[JField]:
        for field in self._tuple_fields:
            if field.foreign_class is fcls:
                if fname is not None:
                    if field.name == fname:
                        return field
                elif fkey is not None:
                    if field.fdef.foreign_key == fkey:
                        return field
        else:
            return None

    def _resolve_ref_types_if_needed(self: CDef) -> None:
        if self._ref_types_resolved is False:
            self._resolve_types()
            self._ref_types_resolved = True

    def _resolve_ref_names_if_needed(self: CDef) -> None:
        if self._ref_names_resolved is False:
            self._resolve_ref_names()
            self._ref_names_resolved = True

    def _resolve_types(self: CDef) -> None:
        for jfield in self._tuple_fields:
            if jfield.types.fdef._unresolved:
                cgraph = self.jconf.cgraph
                jfield._types = rnamedtypes(jfield.types, cgraph, self.name)

    def _resolve_ref_names(self: CDef) -> None:
        for jfield in self._tuple_fields:
            if jfield.types.fdef._fstore == FStore.LOCAL_KEY:
                ref_name_strategy = self.jconf.ref_name_strategy
                if jfield.fdef.ftype == FType.INSTANCE:
                    self._reference_names.append(ref_name_strategy(jfield))
                    self._camelized_reference_names.append(
                        self.jconf.output_key_strategy(ref_name_strategy(jfield)))
                    self._rfmap[ref_name_strategy(jfield)] = jfield
                elif jfield.fdef.ftype == FType.LIST:
                    self._list_reference_names.append(ref_name_strategy(jfield))
                    self._camelized_list_reference_names.append(
                        self.jconf.output_key_strategy(ref_name_strategy(jfield)))
                    self._rfmap[ref_name_strategy(jfield)] = jfield
            elif jfield.types.fdef._fstore == FStore.FOREIGN_KEY:
                if jfield.types.fdef._use_join_table:
                    rkes = self.jconf.ref_name_strategy
                    jkes = self.jconf.output_key_strategy
                    rk = rkes(jfield)
                    self._virtual_reference_names.append(rk)
                    self._camelized_virtual_reference_names.append(jkes(rk))
                    self._virtual_reference_fields[rk] = jfield
                    self._rfmap[rk] = jfield
        self._available_names: set[str] = set(self._field_names
                                              + self._camelized_field_names
                                              + self._reference_names
                                              + self._camelized_reference_names
                                              + self._list_reference_names
                                              + self._camelized_list_reference_names)
        self._update_names: set[str] = set(self._field_names
                                           + self._reference_names
                                           + self._list_reference_names)



"""
This module contains `JSONEncoder`, the encoder class for Python `json` module
that encodes JSON Classes objects.
"""
from typing import Any
from json.encoder import JSONEncoder as PythonDefaultJSONEncoder


class JSONEncoder(PythonDefaultJSONEncoder):
    """The JSONEncoder that works with jsonclasses objects. To dump an json
    encodable object with jsonclasses, use the system default json like this:

      from json import dumps
      from jsonclasses import JSONEncoder

      dumps(obj, cls=JSONEncoder)
    """

    def default(self, o: Any):
        return o.tojson() if hasattr(o.__class__, '__is_jsonclass__') \
            else super().default(o)



from __future__ import annotations
from tests.classes.super_iterable import SuperIterable
from unittest import TestCase


class TestSplit(TestCase):

    def test_split_keeps_list_value_if_it_is_list_of_strs(self):
        s = SuperIterable(itssp=["a", "d", "s", "r"])
        self.assertEqual(s.itssp, ["a", "d", "s", "r"])

    def test_split_splits_str_into_a_list_of_substrs(self):
        s = SuperIterable(itssp="abc.ous.fga.ssr")
        self.assertEqual(s.itssp, ["abc", "ous", "fga", "ssr"])

    def test_split_keeps_callable_value_if_it_is_list_of_strs(self):
        s = SuperIterable(c_itssp=["a", "d", "s", "r"])
        self.assertEqual(s.c_itssp, ["a", "d", "s", "r"])

    def test_split_splits_types_a_list_of_substrs(self):
        s = SuperIterable(t_itssp="abc.ous.fga.ssr")
        self.assertEqual(s.t_itssp, ["abc", "ous", "fga", "ssr"])



"""This module defines ORM protocols, the protocols that objects of JSONClasses
ORM integrations should implement.
"""
from __future__ import annotations
from typing import (
    Protocol, Optional, TypeVar, Union, Any, Generator, Iterator
)
from .jobject import JObject


T = TypeVar('T', bound='ORMObject', covariant=True)
U = TypeVar('U', bound='BaseQuery')
V = TypeVar('V', bound='BaseListQuery')


class Subquery(Protocol[U]):
    name: str
    query: Optional[U]


class BaseQuery(Protocol[T]):

    def __init__(self: U, cls: type[T]) -> None:
        ...

    def include(self: U, name: str, query: Optional[BaseQuery] = None) -> U:
        ...


class BaseListQuery(BaseQuery[T]):

    def __init__(self: V,
                 cls: type[T],
                 filter: Union[dict[str, Any], str, None] = None) -> None:
        ...

    def order(self: V, field: str, sort: Optional[int] = None) -> V:
        ...

    def skip(self: V, n: int) -> V:
        ...

    def limit(self: V, n: int) -> V:
        ...

    def page_number(self: V, n: int) -> V:
        ...

    def page_no(self: V, n: int) -> V:
        ...

    def page_size(self: V, n: int) -> V:
        ...

    def pick(self: V, fields: list[str]) -> V:
        ...

    def omit(self: V, fields: list[str]) -> V:
        ...


class ListQuery(BaseListQuery[T]):

    def exec(self) -> list[T]:
        ...

    def __await__(self) -> Generator[None, None, list[T]]:
        ...


class SingleQuery(BaseListQuery[T]):

    def exec(self) -> T:
        ...

    def __await__(self) -> Generator[None, None, T]:
        ...

    @property
    def optional(self) -> OptionalSingleQuery:
        ...


class OptionalSingleQuery(BaseListQuery[T]):

    def exec(self) -> Optional[T]:
        ...

    def __await__(self) -> Generator[None, None, Optional[T]]:
        ...



class BaseIDQuery(BaseQuery[T]):

    def __init__(self: U, cls: type[T], id: Union[str, int]) -> None:
        ...


class IDQuery(BaseIDQuery[T]):

    def exec(self) -> T:
        ...

    def __await__(self) -> Generator[None, None, T]:
        ...

    @property
    def optional(self) -> OptionalIDQuery:
        ...


class OptionalIDQuery(BaseIDQuery[T]):

    def exec(self) -> Optional[T]:
        ...

    def __await__(self) -> Generator[None, None, Optional[T]]:
        ...


class IDSQuery(BaseQuery[T]):

    def exec(self) -> list[T]:
        ...

    def __await__(self) -> Generator[None, None, list[T]]:
        ...


class ExistQuery(BaseListQuery[T]):

    def exec(self) -> bool:
        ...

    def __await__(self) -> Generator[None, None, bool]:
        ...


class IterateQuery(BaseListQuery[T]):

    def exec(self) -> Iterator[T]:
        ...

    def __await__(self) -> Generator[None, None, Iterator[T]]:
        ...


class ORMObject(JObject):
    """The `ORMObject` protocol defines methods that JSONClasses ORM
    integrations should implement.
    """


    @classmethod
    def find(cls: type[T], *args, **kwargs: Any) -> ListQuery[T]:
        ...

    @classmethod
    def one(cls: type[T], *args, **kwargs: Any) -> SingleQuery[T]:
        ...

    @classmethod
    def id(cls: type[T], id: Any, *args, **kwargs) -> IDQuery[T]:
        ...

    @classmethod
    def ids(cls: type[T], ids: list[Any], *args, **kwargs) -> IDSQuery[T]:
        ...

    @classmethod
    def linked(cls: type[T], *args, **kwargs: Any) -> BaseQuery[T]:
        ...

    @classmethod
    def exist(cls: type[T], **kwargs: Any) -> ExistQuery[T]:
        ...

    @classmethod
    def iterate(cls: type[T], **kwargs: Any) -> IterateQuery[T]:
        ...

    def _orm_delete(self: T, no_raise: bool = False) -> None:
        ...

    def _orm_restore(self: T) -> None:
        ...



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_oneof import SuperOneOf


class TestOneOf(TestCase):

    def test_oneof_wont_raise_with_constant(self):
        SuperOneOf(valconst='abc').validate()

    def test_oneof_raises_with_constant(self):
        self.assertRaises(ValidationException, SuperOneOf(valconst='qwe').validate)

    def test_oneof_wont_raise_with_callable(self):
        SuperOneOf(valcallable='abc').validate()

    def test_oneof_raises_with_callable(self):
        self.assertRaises(ValidationException, SuperOneOf(valcallable='qwe').validate)

    def test_oneof_wont_raise_with_types(self):
        SuperOneOf(valtypes='abc').validate()

    def test_oneof_raises_with_types(self):
        self.assertRaises(ValidationException, SuperOneOf(valtypes='qwe').validate)



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.product_name import AlphaProductName


class TestIsAlpha(TestCase):

    def test_alpha_doesnt_raise_if_value_is_alpha(self):
        analysis = AlphaProductName(product_name='water')
        analysis.validate()

    def test_alpha_raises_if_value_is_number(self):
        analysis = AlphaProductName(product_name='123')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productName'],
                         "value is not alpha str")

    def test_alpha_raises_if_value_contains_number(self):
        analysis = AlphaProductName(product_name='water12')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productName'],
                         "value is not alpha str")

    def test_alpha_raises_if_value_contains_special_character(self):
        analysis = AlphaProductName(product_name='water!')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['productName'],
                         "value is not alpha str")



"""module for reverse modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ReverseModifier(Modifier):
    """Reverse modifier reverses iterable values."""

    def transform(self, ctx: Ctx) -> Any:
        is_iterable = type(ctx.val) is str or isinstance(ctx.val, list)
        return ctx.val[::-1] if is_iterable else ctx.val



from __future__ import annotations
from tests.classes.simple_math import SimpleMath
from unittest import TestCase

class TestPow(TestCase):

    def test_pow_with_get_the_value_of_original_value_to_the_power_of_int_value(self):
        p = SimpleMath(i_pow=3)
        self.assertEqual(p.i_pow, 8)

    def test_pow_with_get_the_value_of_original_value_to_the_power_of_float_value(self):
        p = SimpleMath(f_pow=3)
        self.assertEqual(p.f_pow, 15.625)

    def test_pow_with_get_the_value_of_callable_value_to_the_power_of_float_value(self):
        p = SimpleMath(c_pow=3)
        self.assertEqual(p.c_pow, 15.625)

    def test_pow_with_get_the_value_of_types_value_to_the_power_of_float_value(self):
        p = SimpleMath(c_pow=3)
        self.assertEqual(p.c_pow, 15.625)



from __future__ import annotations
from unittest import TestCase
from tests.classes.role_user import RoleUser


class TestReadonlyModifier(TestCase):

    def test_readonly_will_not_be_set_through_init(self):
        user = RoleUser(username='John', role='I want to change haha XD')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'normal'})

    def test_readonly_will_not_override_default_value_through_init(self):
        user = RoleUser(username='John', role='I want to change haha XD')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'normal'})

    def test_readonly_will_be_ok_when_not_provided(self):
        user = RoleUser(username='John')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'normal'})

    def test_readonly_will_not_be_set_through_set(self):
        user = RoleUser()
        user.set(username='John', role='I want to change haha XD')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'normal'})

    def test_readonly_will_not_be_erased_through_set(self):
        user = RoleUser()
        user.set(username='John', role='I want to change haha XD')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'normal'})

    def test_readonly_can_be_modified_directly(self):
        user = RoleUser()
        user.set(username='John', role='I want to change haha XD')
        user.role = 'designer'
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'designer'})

    def test_readonly_can_be_updated_through_update(self):
        user = RoleUser()
        user.set(username='John', role='I want to change haha XD')
        user.update(role='admin')
        self.assertEqual(user._data_dict,
                         {'username': 'John', 'role': 'admin'})



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_order import SimpleOrder


class TestIsNew(TestCase):

    def test_jobject_is_new_on_create(self):
        order = SimpleOrder(quantity=5)
        self.assertEqual(order.is_new, True)

    def test_jobject_is_new_is_readonly(self):
        order = SimpleOrder(quantity=5)
        with self.assertRaises(AttributeError) as context:
            order.is_new = False
        self.assertEqual(str(context.exception), "can't set attribute")

    def test_jobject_new_object_wont_record_modified_fields(self):
        order = SimpleOrder(quantity=5)
        order.quantity = 2
        order.quantity = 10
        self.assertEqual(order.modified_fields, ())



"""module for hassuffix modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class IsSuffixOfModifier(Modifier):
    """Has suffix modifier for checking if a str is suffix of another str."""

    def __init__(self, suffix: str | list[Any] | Callable | Types) -> None:
        self.suffix = suffix

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, list):
            if not self.resolve_param(self.suffix, ctx)[-len(ctx.val):] == ctx.val:
                ctx.raise_vexc('suffix is not found')
        elif type(ctx.val) is str:
            if not self.resolve_param(self.suffix, ctx).endswith(ctx.val):
                ctx.raise_vexc('suffix is not found')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_order import SimpleOrder


class TestInt(TestCase):

    def test_int_is_int_after_assigned(self):
        order = SimpleOrder(quantity=5)
        self.assertEqual(order._data_dict, {'name': None, 'quantity': 5})

    def test_int_raises_if_value_is_not_int(self):
        order = SimpleOrder(quantity=10.5)
        with self.assertRaises(ValidationException) as context:
            order.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['quantity'],
                         "value is not int")

    def test_int_is_int_when_tojson(self):
        order = SimpleOrder(quantity=5)
        self.assertEqual(order.tojson(), {'quantity': 5})



"""module for neq modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class NeqModifier(Modifier):
    """Neq modifier validates value by unequal testing."""

    def __init__(self, val: Any | Types | Callable):
        self.val = val

    def validate(self, ctx: Ctx) -> None:
        if ctx.val == self.resolve_param(self.val, ctx):
            ctx.raise_vexc('value is not unequal')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types

@jsonclass
class SuperRandom:
    random_digits: Optional[int] = types.int.randomdigits(1)
    random_alnums: Optional[int] = types.int.randomalnums(1)
    random_alnumpuncs: Optional[int] = types.int.randomalnumpuncs(1)
    random_int: Optional[int] = types.int.randomint(10, 11)
    random_float: Optional[float] = types.float.randomfloat(10.0, 10.0)

    random_cfloat: Optional[float] = types.float.randomfloat(lambda: 10.0, lambda: 10.0)
    random_tfloat: Optional[float] = types.float.randomfloat(types.default(10.0), types.default(10.0))
    c_random_int: Optional[int] = types.int.randomint(lambda: 10, lambda: 11)
    t_random_int: Optional[int] = types.int.randomint(types.default(10), types.default(11))
    c_random_digits: Optional[int] = types.int.randomdigits(lambda: 1)
    t_random_digits: Optional[int] = types.int.randomdigits(types.default(1))



from __future__ import annotations
from jsonclasses import jsonclass, types


def create_first_revision(article: CBOArticle, operator: str) -> None:
    CBORevision(name=operator).article = article


def force_content(article: CBOArticle, operator: str) -> None:
    article.content = operator


@jsonclass
class CBORevision:
    name: str
    article: CBOArticle = types.objof('CBOArticle').linkto.required


@jsonclass(on_create=[create_first_revision, force_content])
class CBOArticle:
    name: str
    content: str
    revisions: list[CBORevision] = types.nonnull.listof('CBORevision') \
                                        .linkedby('article')



from unittest import TestCase
from tests.classes.gender_user import GenderUser


class TestWriteonce(TestCase):

    def test_writeonce_cant_be_updated_by_set_if_its_value_is_present(self):
        user = GenderUser(nickname='Big Cock Brother', gender='male')
        user.set(**{'nickname': '20cm Brother', 'gender': 'female'})
        self.assertEqual(user.nickname, '20cm Brother')
        self.assertEqual(user.gender, 'male')

    def test_writeonce_can_be_updated_by_set_if_its_value_is_not_present(self):
        user = GenderUser(nickname='Alice')
        user.set(**{'nickname': 'Alice Heart', 'gender': 'female'})
        self.assertEqual(user.nickname, 'Alice Heart')
        self.assertEqual(user.gender, 'female')

    def test_writeonce_can_be_updated_by_update_anyway(self):
        user = GenderUser(nickname='Alice', gender='male')
        user.update(**{'nickname': 'Alice Heart', 'gender': 'female'})
        self.assertEqual(user.nickname, 'Alice Heart')
        self.assertEqual(user.gender, 'female')

    def test_writeonce_can_be_updated_directly_anyway(self):
        user = GenderUser(nickname='Alice', gender='male')
        user.gender = 'female'
        self.assertEqual(user.gender, 'female')



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class NewUnion:
    name: str | None
    code: str | int | None


@jsonclass
class NewRUnion:
    code: str | int
    items: dict[str, str | int] | list[str]


@jsonclass
class NewOUnion:
    items: dict[str, str | int] | list[str] | None



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from inspect import signature
from ..fdef import FStore, FDef
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class SetterModifier(Modifier):
    """Setter modifier provides setter to a calculated field."""

    def __init__(self, calc: Callable | Types) -> None:
        from ..types import Types
        self.calc = calc
        if not isinstance(calc, Types):
            if not callable(calc):
                raise ValueError('setter is not callable')
            params_len = len(signature(calc).parameters)
            if params_len != 2:
                raise ValueError('not a valid setter')

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.CALCULATED
        fdef._setter = self.calc



from __future__ import annotations
from tests.classes.super_iterable import SuperIterable
from unittest import TestCase



class TestReverse(TestCase):

    def test_reverse_reverses_str(self):
        s = SuperIterable(itsr="Hello Ben")
        self.assertEqual(s.itsr, "neB olleH")

    def test_reverse_reverses_list(self):
        s = SuperIterable(itl=["Hello Ben", 3, True, 6, False])
        self.assertEqual(s.itl, [False, 6, True, 3, "Hello Ben"])



"""module for round modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class RoundModifier(Modifier):
    """Round modifier rounds number value."""

    def transform(self, ctx: Ctx) -> Any:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        return round(ctx.val) if is_number else ctx.val



"""module for unqueryable modifier."""
from ..fdef import FDef, Queryability
from .modifier import Modifier


class UnqueryableModifier(Modifier):
    """Unqueryable modifier marks a column should be unqueryable."""

    def define(self, fdef: FDef) -> None:
        fdef._queryability = Queryability.UNQUERYABLE



from __future__ import annotations
from unittest import TestCase
from tests.classes.user_sos import UserSOS, UserSOSZ
from tests.classes.linked_sos import (LinkedUserSOS, LinkedBookSOS,
                                      LinkedUserSOSE, LinkedBookSOSE)


class TestSetOnSave(TestCase):

    def test_new_object_triggers_setonsave(self):
        user = UserSOS(name='n', age=50)
        user._set_on_save()
        self.assertEqual(user.age, 350)

    def test_modified_object_triggers_setonsave(self):
        user = UserSOS(name='n', age=100)
        setattr(user, '_is_new', False)
        user.name = 'N'
        user._set_on_save()
        self.assertEqual(user.age, 400)

    def test_unmodified_object_wont_trigger_setonsave(self):
        user = UserSOS(name='n', age=100)
        setattr(user, '_is_new', False)
        user._set_on_save()
        self.assertEqual(user.age, 100)

    def test_setonsave_accepts_0_args(self):
        user = UserSOSZ(name='n', age=100)
        setattr(user, '_is_new', False)
        user.name = 'N'
        user._set_on_save()
        self.assertEqual(user.age, 500)

    def test_setonsave_triggers_for_modified_linked_objects(self):
        book1 = LinkedBookSOS(id=1, name='B1', value=1)
        book2 = LinkedBookSOS(id=2, name='B2', value=1)
        setattr(book1, '_is_new', False)
        setattr(book2, '_is_new', False)
        user = LinkedUserSOS(id=1, name='U', value=1)
        setattr(user, '_is_new', False)
        user.books.append(book1)
        user.books.append(book2)
        user._set_on_save()
        self.assertEqual(user.value, 2)
        self.assertEqual(book1.value, 2)
        self.assertEqual(book2.value, 2)

    def test_setonsave_triggers_even_root_is_unmodified(self):
        book1 = LinkedBookSOS(id=1, name='B1', value=1)
        book2 = LinkedBookSOS(id=2, name='B2', value=1)
        setattr(book1, '_is_new', False)
        setattr(book2, '_is_new', False)
        user = LinkedUserSOS(id=1, name='U', value=1)
        setattr(user, '_is_new', False)
        user.books.append(book1)
        user.books.append(book2)
        setattr(user, '_is_modified', False)
        setattr(user, '_modified_fields', set())
        user._set_on_save()
        self.assertEqual(user.value, 1)
        self.assertEqual(book1.value, 2)
        self.assertEqual(book2.value, 2)

    def test_setonsave_triggers_anyway_if_owner_modified_for_embedded(self):
        book1 = LinkedBookSOSE(id=1, name='B1', value=1)
        book2 = LinkedBookSOSE(id=2, name='B2', value=1)
        setattr(book1, '_is_new', False)
        setattr(book2, '_is_new', False)
        user = LinkedUserSOSE(id=1, name='U', value=1, books=[book1, book2])
        book1.name = 'BB1'
        book2.name = 'BB2'
        setattr(user, '_is_modified', False)
        setattr(user, '_modified_fields', set())
        user._set_on_save()
        self.assertEqual(user.value, 2)
        self.assertEqual(book1.value, 2)
        self.assertEqual(book2.value, 2)

    def test_setonsave_triggers_anyway_if_owner_unmodified_for_embedded(self):
        book1 = LinkedBookSOSE(id=1, name='B1', value=1)
        book2 = LinkedBookSOSE(id=2, name='B2', value=1)
        setattr(book1, '_is_new', False)
        setattr(book2, '_is_new', False)
        user = LinkedUserSOSE(id=1, name='U', value=1, books=[book1, book2])
        book1.name = 'BB1'
        book2.name = 'BB2'
        setattr(user, '_is_new', False)
        setattr(user, '_is_modified', False)
        setattr(user, '_modified_fields', set())
        user._set_on_save()
        self.assertEqual(user.value, 1)
        self.assertEqual(book1.value, 2)
        self.assertEqual(book2.value, 2)



"""module for fval modifier."""
from __future__ import annotations
from typing import Any, Callable, cast, TYPE_CHECKING
from ..objref import ObjRef
from ..jobject import JObject
from ..fdef import FStore
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class FValModifier(Modifier):
    """Get value at field from a JSONClass object.
    """

    def __init__(self, field_name: str | Callable | Types) -> None:
        self.field_name = field_name

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        field_name = self.resolve_param(self.field_name, ctx)
        val = cast(JObject, ctx.val)
        field = val.__class__.cdef.field_named(field_name)
        obj = getattr(val, field_name)
        if field.fdef.fstore == FStore.LOCAL_KEY:
            if obj is not None:
                return obj
            else:
                kt = val.__class__.cdef.jconf.ref_name_strategy
                fidname = kt(field)
                cls_name = field.fdef.inst_cls.__name__
                ref_id = getattr(val, fidname)
                if ref_id is None:
                    return None
                else:
                    return ObjRef(cls=cls_name, id=ref_id)
        else:
            return obj



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_company import SimpleCompany
from tests.classes.simple_code import SimpleCode


class TestLength(TestCase):

    def test_length_does_not_raise_if_length_is_in_between(self):
        c1 = SimpleCompany(name='AZ', number_of_employees=5)
        c1.validate()
        c2 = SimpleCompany(name='QWERT', number_of_employees=5)
        c2.validate()
        c3 = SimpleCompany(name='QWERTQWERT', number_of_employees=5)
        c3.validate()

    def test_length_raises_if_length_is_lt_lowerbond(self):
        c1 = SimpleCompany(name='Q', number_of_employees=5)
        with self.assertRaises(ValidationException):
            c1.validate()

    def test_length_raises_if_length_is_gt_lowerbond(self):
        c1 = SimpleCompany(name='QWERTYUIOP{', number_of_employees=5)
        with self.assertRaises(ValidationException):
            c1.validate()

    def test_length_should_match_length_in_str(self):
        code = SimpleCode(code='1234')
        code.validate()

    def test_length_should_match_length_in_list(self):
        l_code = SimpleCode(l_code=[1, 2, 3, 4])
        l_code.validate()

    def test_length_raises_if_length_does_not_match_in_str(self):
        code = SimpleCode(code='12345')
        with self.assertRaises(ValidationException):
            code.validate()
        code = SimpleCode(code='1')
        with self.assertRaises(ValidationException):
            code.validate()

    def test_length_raises_if_length_does_not_match_in_list(self):
        l_code = SimpleCode(l_code=[1, 2, 3, 4, 5, 0])
        with self.assertRaises(ValidationException):
            l_code.validate()
        l_code = SimpleCode(l_code=[1, 2])
        with self.assertRaises(ValidationException):
            l_code.validate()

    def test_length_accepts_single_callable_param(self):
        c_code = SimpleCode(c_code='1234')
        c_code.validate()
        c_code = SimpleCode(c_code='12345')
        with self.assertRaises(ValidationException):
            c_code.validate()

    def test_length_accepts_double_callable_param(self):
        cd_code = SimpleCode(cd_code='1234')
        cd_code.validate()
        cd_code = SimpleCode(cd_code='12345')
        cd_code.validate()
        cd_code = SimpleCode(cd_code='123456')
        with self.assertRaises(ValidationException):
            cd_code.validate()
        cd_code = SimpleCode(cd_code='123')
        with self.assertRaises(ValidationException):
            cd_code.validate()

    def test_length_accepts_single_types_param(self):
        t_code = SimpleCode(t_code='1234')
        t_code.validate()
        t_code = SimpleCode(t_code='12345')
        with self.assertRaises(ValidationException):
            t_code.validate()

    def test_length_accepts_double_types_param(self):
        td_code = SimpleCode(td_code='1234')
        td_code.validate()
        td_code = SimpleCode(td_code='12345')
        td_code.validate()
        td_code = SimpleCode(td_code='123456')
        with self.assertRaises(ValidationException):
            td_code.validate()
        td_code = SimpleCode(td_code='123')
        with self.assertRaises(ValidationException):
            td_code.validate()



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_int import SuperInt


class TestOdd(TestCase):

    def test_odd_doesnt_raise_if_int_value_is_odd(self):
        i = SuperInt(i_o=5)
        self.assertEqual(i.i_o, 5)

    def test_odd_raises_if_int_value_is_not_odd(self):
        i = SuperInt(i_o=6)
        with self.assertRaises(ValidationException) as context:
            i.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['iO'],
                         "value is not odd")



"""This module defines utility functions for working with keypaths."""
from __future__ import annotations
from typing import Any, Union, TYPE_CHECKING
from re import split,search
from inflection import camelize, singularize, underscore
from .fdef import FStore, FType
if TYPE_CHECKING:
    from .jfield import JField


def camelize_key(key: str) -> str:
    return camelize(key)


def underscore_key(key: str) -> str:
    return underscore(key)


def identical_key(key: str) -> str:
    return key


def reference_key(field: JField) -> str:
    """
    Figure out the correct reference key name from the field definition.

    Args:
        field (JField): The JSON class field to figure out reference
        key from.

    Returns:
        str: The reference key which represents this field.

    Raises:
        ValueError: ValueError is raised if the field definition is not a \
            supported reference field.
    """
    if field.fdef.fstore not in \
            [FStore.FOREIGN_KEY, FStore.LOCAL_KEY]:
        raise ValueError(f"field named {field.name} is not a reference field")
    if field.fdef.ftype == FType.LIST:
        return singularize(field.name) + '_ids'
    elif field.fdef.ftype == FType.INSTANCE:
        return field.name + '_id'
    else:
        raise ValueError(f"field type {field.fdef.ftype} is not a "
                         "supported reference field type")


def concat_keypath(*args: Union[str, int]) -> str:
    """Concatenate partial keypaths and keys into a concatenated single
    keypath.

    Args:
        *args (Union[str, int]): The partial keypaths and keys to concatenate.

    Returns:
        str: the concatenated keypath.
    """
    return '.'.join([str(arg) for arg in args if len(str(arg)) > 0])


def keypath_drop_last(keypath: str) -> str:
    """Drop the last part of a keypath. If it only has one part, empty string
    is returned. If it's empty string, empty string is returned.

    Args:
        keypath (str): The keypath to drop last from.

    Returns:
        str: A new keypath with last component dropped or empty string.
    """
    return '.'.join(keypath.split('.')[:-1])


def initial_keypaths(keypaths: set[str]) -> set[str]:
    """Get a set of initial keypath component from `keypaths`.

    Args:
        keypaths (set[str]): A set of keypaths.

    Returns:
        set[str]: A set of initial keypath components without duplication.
    """
    return set([initial_keypath(k) for k in keypaths])


def initial_keypath(keypath: str) -> str:
    """Get the initial keypath component from the keypath.

    Args:
        keypath (str): The keypath to fetch the initial component from.

    Returns:
        str: The initial keypath component or empty string.
    """
    return keypath.split('.')[0]


def single_key_args(kwargs: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if not search(r'\.|\[|\]', k)}


def compound_key_args(kwargs: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if search(r'\.|\[|\]', k)}


def keypath_split(key: str) -> list[str]:
    return [k for k in split(r'\]|\[|\.', key) if len(k) > 0]



from __future__ import annotations
from typing import Optional, Any
from jsonclasses import jsonclass, types


@jsonclass
class SuperType:
    tl: Optional[list[Any]] = types.tolist.listof(Any)
    tb: Optional[bool] = types.tobool.bool
    ti: Optional[int] = types.toint.int
    tf: Optional[float] = types.tofloat.float
    ts: Optional[str] = types.tostr.str



"""module for modifier modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class InvalidModifier(Modifier):
    """A modifier that turns value into invalid."""

    def validate(self, ctx: Ctx) -> None:
        """Raises invalid exception."""
        ctx.raise_vexc('value is invalid')



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase

from tests.classes.super_datetime import SuperDateTime


class TestTobomon(TestCase):

    def test_tobomon_transforms_datetime_into_the_time_of_beginning_of_month(self):
        d = SuperDateTime(dtbmth=datetime(2021, 10, 11, 17, 37, 27,43235))
        self.assertEqual(d.dtbmth, datetime(2021,10, 1, 0, 0))

    def test_tobomon_transforms_date_into_the_time_of_beginning_of_month(self):
        d = SuperDateTime(dbmth=date(2021, 10, 11))
        self.assertEqual(d.dbmth, date(2021,10, 1))

    def test_tobomon_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(sbmth="12345")
        self.assertEqual(s.sbmth, "12345")



from __future__ import annotations
from typing import Final


linkto: Final[str] = 'linkto'


def linkedby(key: str) -> str:
    return f"linkedby('{key}')"


def linkedthru(key: str) -> str:
    return f"linkedthru('{key}')"



"""module for float modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any
from .type_modifier import TypeModifier
from ..fdef import FType
if TYPE_CHECKING:
    from ..ctx import Ctx


class FloatModifier(TypeModifier):
    """Date modifier validate value against float type."""

    def __init__(self):
        super().__init__()
        self.cls = float
        self.ftype = FType.FLOAT

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if type(ctx.val) is int:
            return float(ctx.val)
        return ctx.val
